import pandas as pd 
import numpy as np 
from torchkeras.pbar import is_jupyter 

class VLogCallback:
    def __init__(self, monitor_metric='val_Accuracy', monitor_mode='max', figsize = (8,6), save_path='history.png'):
        import matplotlib.pyplot as plt
        self.in_jupyter = is_jupyter()
        self.plt = plt
        self.figsize = figsize
        self.metric = monitor_metric
        self.mode = monitor_mode
        self.save_path = save_path
        x_bounds = [0, 10]
        title = f'best {self.metric} = ?'
        self.update_graph(None, title=title, x_bounds = x_bounds)
        
    def get_dfhistory(self, info):
        from copy import deepcopy
        if not hasattr(info,'metrics'):
            return pd.DataFrame() 
        
        dic = deepcopy(info.metrics) 
        if 'learn' in dic:
            dic['train'] = dic['learn']
            dic.pop('learn')

        if 'validation' in dic:
            dic['val'] = dic['validation']
            dic.pop('validation')

        dfhis_train = pd.DataFrame(dic['train']) 
        dfhis_train.columns = ['train_'+x for x in dfhis_train.columns]

        dfhis_val = pd.DataFrame(dic['val']) 
        dfhis_val.columns = ['val_'+x for x in dfhis_val.columns]
        dfhistory = dfhis_train.join(dfhis_val)
        dfhistory['iteration'] = range(1,len(dfhistory)+1)
        
        return dfhistory 

    def after_iteration(self, info):
        dfhistory = self.get_dfhistory(info)
        n = len(dfhistory)
        x_bounds = [dfhistory['iteration'].min(), 10+(n//10)*10]
        title = self.get_title(info)
        self.update_graph(info, title = title, x_bounds = x_bounds)
        return True
        
    def get_best_score(self, info):
        dfhistory = self.get_dfhistory(info)
        arr_scores = dfhistory[self.metric]
        best_score = np.max(arr_scores) if self.mode=="max" else np.min(arr_scores)
        best_iteration = dfhistory.loc[arr_scores==best_score,'iteration'].tolist()[0]
        return (best_iteration, best_score)
        
    def get_title(self,  info):
        best_iteration,best_score = self.get_best_score(info)
        title = f'best {self.metric} = {best_score:.4f} (@iteration {best_iteration})'
        return title

    def update_graph(self, info, title=None, x_bounds=None, y_bounds=None):
        from IPython.display  import display 
        if not hasattr(self, 'graph_fig'):
            self.graph_fig, self.graph_ax = self.plt.subplots(1, figsize=self.figsize)
            self.graph_out = display(self.graph_ax.figure, display_id=True)
        self.graph_ax.clear()
        dfhistory = self.get_dfhistory(info)
        iterations = dfhistory['iteration'] if 'iteration' in dfhistory.columns else []
        
        m1 = "train_"+self.metric.replace('val_','')
        if  m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            
            self.graph_ax.plot(iterations,train_metrics,'bo--',label= m1)

        m2 = 'val_'+self.metric.replace('val_','')
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.graph_ax.plot(iterations,val_metrics,'co-',label =m2)


        self.graph_ax.set_xlabel("iteration")
        self.graph_ax.set_ylabel(self.metric.replace('val_',''))  
        if title:
             self.graph_ax.set_title(title)
        if m1 in dfhistory.columns or m2 in dfhistory.columns or self.metric in dfhistory.columns:
            self.graph_ax.legend(loc='best')
            
        if len(iterations)>0:
            best_iteration, best_score = self.get_best_score(info)
            self.graph_ax.plot(best_iteration,best_score,'r*',markersize=15)

        if x_bounds is not None: self.graph_ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.graph_ax.set_ylim(*y_bounds)
        
        if self.in_jupyter:
            self.graph_out.update(self.graph_ax.figure)
        self.graph_fig.savefig(self.save_path)
        
        self.plt.close();
        
if __name__ =='__main__':
    from IPython.display import display 

    import datetime,json
    import numpy as np
    import pandas as pd
    import catboost as cb 
    from catboost.datasets import titanic
    from sklearn import datasets
    from sklearn.model_selection import train_test_split
    from sklearn.model_selection import StratifiedKFold

    from sklearn.metrics import f1_score,roc_auc_score,accuracy_score
    import plotly.graph_objs as go 
    import plotly.express as px 

    from torchkeras.tools.catboost import VLogCallback 

    def printlog(info):
        nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        print("\n"+"=========="*8 + "%s"%nowtime)
        print(info+'...\n\n')


    #================================================================================
    printlog("step1: preparing data...")

    dfdata,dftest = titanic()

    display(dfdata.head()) 

    label_col = "Survived"

    dfnull = pd.DataFrame(dfdata.isnull().sum(axis=0),columns = ["null_cnt"]).query("null_cnt>0")
    print("null_features:") 
    print(dfnull)

    dfdata.fillna(-9999, inplace=True)
    dftest.fillna(-9999, inplace=True)


    cate_cols = [x for x in dfdata.columns 
                 if dfdata[x].dtype not in [np.float32,np.float64] and x!=label_col]
    for col in cate_cols:
        dfdata[col] = pd.Categorical(dfdata[col]) 
        dftest[col] = pd.Categorical(dftest[col]) 

    dftrain,dfvalid = train_test_split(dfdata, train_size=0.75, random_state=42)
    Xtrain,Ytrain = dftrain.drop(label_col,axis = 1),dftrain[label_col]
    Xvalid,Yvalid = dfvalid.drop(label_col,axis = 1),dfvalid[label_col]
    cate_cols_indexs = np.where(Xtrain.columns.isin(cate_cols))[0]


    pool_train = cb.Pool(data = Xtrain, label = Ytrain, cat_features=cate_cols)
    pool_valid = cb.Pool(data = Xvalid, label = Yvalid, cat_features=cate_cols)



    #================================================================================
    printlog("step2: setting parameters...")

    iterations = 1000
    early_stopping_rounds = 200

    params = {
        'learning_rate': 0.05,
        'loss_function': "Logloss",
        'eval_metric': "Accuracy",
        'depth': 6,
        'min_data_in_leaf': 20,
        'random_seed': 42,
        'logging_level': 'Silent',
        'use_best_model': True,
        'one_hot_max_size': 5,  
        'boosting_type':"Ordered", 
        'max_ctr_complexity': 2, 
        'nan_mode': 'Min' 
    }


    #================================================================================
    printlog("step3: training model...")


    model = cb.CatBoostClassifier(
        iterations = iterations,
        early_stopping_rounds = early_stopping_rounds,
        train_dir='catboost_info/',
        **params
    )

    vlog_cb = VLogCallback(monitor_metric='val_Accuracy',monitor_mode='max')

    model.fit(
        pool_train,
        eval_set=pool_valid,
        plot=False,
        callbacks=[vlog_cb]
    )


    #================================================================================
    printlog("step4: evaluating model ...")


    y_pred_train = model.predict(Xtrain)
    y_pred_valid = model.predict(Xvalid)

    train_score = f1_score(Ytrain,y_pred_train)
    valid_score = f1_score(Yvalid,y_pred_valid)


    print('train f1_score: {:.5} '.format(train_score))
    print('valid f1_score: {:.5} \n'.format(valid_score))   


    #feature importance 
    dfimportance = model.get_feature_importance(prettified=True) 
    dfimportance = dfimportance.sort_values(by = "Importances").iloc[-20:]
    fig_importance = px.bar(dfimportance,x="Importances",y="Feature Id",title="Feature Importance")

    fig_importance.show() 

    #score distribution
    y_test_prob = model.predict_proba(dfvalid.drop(label_col,axis = 1))[:,-1]
    fig_hist = px.histogram(
        x=y_test_prob,color =dfvalid[label_col],  nbins=50,
        title = "Score Distribution",
        labels=dict(color='True Labels', x='Score')
    )
    fig_hist.show() 


    #================================================================================
    printlog("step5: using model ...")

    y_pred_test = model.predict(dftest)
    y_pred_test_prob = model.predict_proba(dftest)


    #================================================================================
    printlog("step6: saving model ...")

    model_dir = 'catboost_model'
    model.save_model(model_dir)
    model_loaded = cb.CatBoostClassifier()
    model.load_model(model_dir)
        


import numpy as np
import PIL
from PIL import Image,ImageDraw,ImageFont
from pathlib import Path
import sys 

simhei_file = str(Path(__file__).parent/"assets/SimHei.ttf")

"""
@author : lyhue1991
@description : keras plot code
"""

def set_matplotlib_font(font_size=15):
    """
    Set the font style and size for Matplotlib plots.

    Args:
        font_size (int): Font size to set.
    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    try:
        import shutil
        shutil.rmtree(mpl.get_cachedir())
    except:
        pass
    simhei_file = Path(__file__).parent/"assets/SimHei.ttf"
    ttf_dir = Path(mpl.__path__[0])/'mpl-data'/'fonts'/'ttf'
    shutil.copy(str(simhei_file),str(ttf_dir))
    plt.rcParams['font.family'] = ['sans-serif']
    plt.rcParams['font.size'] = str(font_size)
    plt.rcParams['font.sans-serif'] = ['SimHei']
    plt.rcParams['axes.unicode_minus'] = False
    print(f"set matplotlib font to SimHei and fontsize to {font_size}")


def plot_metric(dfhistory, metric):
    """
    Plot training and validation metrics over epochs using Plotly.

    Args:
        dfhistory (pd.DataFrame): DataFrame containing training history.
        metric (str): Name of the metric to plot.

    Returns:
        go.Figure: Plotly figure.
    """
    import plotly.graph_objs as go
    train_metrics = dfhistory["train_"+metric].values.tolist()
    val_metrics = dfhistory['val_'+metric].values.tolist()
    epochs = list(range(1, len(train_metrics) + 1))

    train_scatter = go.Scatter(x=epochs, y=train_metrics, mode="lines+markers",
                              name='train_'+metric, marker=dict(size=8, color="blue"),
                              line=dict(width=2, color="blue", dash="dash"))
    val_scatter = go.Scatter(x=epochs, y=val_metrics, mode="lines+markers",
                            name='val_'+metric, marker=dict(size=10, color="red"),
                            line=dict(width=2, color="red", dash="solid"))
    fig = go.Figure(data=[train_scatter, val_scatter])

    if any([m in metric.lower() for m in ['loss','mse','mae','error','err','l1','l2']]):
        best_metric = np.array(val_metrics).min()
    elif any([m in metric.lower() for m in ['acc','iou','recall','precision','auc']]):
        best_metric = np.array(val_metrics).max()
    elif train_metrics.argmax() > train_metrics.argmin():
        best_metric = np.array(val_metrics).max()
    else:
        best_metric = np.array(val_metrics).min()

    fig.update_layout({"title": "best val_"+metric+"="+f"{best_metric:.4f}",
                       "title_x": 0.45,
                       "xaxis.title": "epoch",
                       "yaxis.title": metric,
                       "font.size": 15,
                       "height": 500,
                       "width": 800
                      })

    return fig


def plot_importance(features, importances, topk=20):
    """
    Plot feature importance using Plotly.

    Args:
        features (list): List of feature names.
        importances (list): List of feature importances.
        topk (int): Number of top features to display.

    Returns:
        go.Figure: Plotly figure.
    """
    import pandas as pd
    import plotly.express as px
    dfimportance = pd.DataFrame({'feature': features, 'importance': importances})
    dfimportance = dfimportance.sort_values(by="importance").iloc[-topk:]
    fig = px.bar(dfimportance, x="importance", y="feature", title="Feature Importance")
    return fig


def plot_score_distribution(labels, scores):
    """
    Plot the score distribution for binary classification problems using Plotly.

    Args:
        labels (list): True labels.
        scores (list): Predicted scores.

    Returns:
        go.Figure: Plotly figure.
    """
    import plotly.express as px
    fig = px.histogram(
        x=scores, color=labels, nbins=50,
        title="Score Distribution",
        labels=dict(color='True Labels', x='Score')
    )
    return fig


def vis_dataframe(df, img_col='img', img_width=100, img_height=100):
    """
    Visualize a DataFrame containing image paths using HTML.

    Args:
        df (pd.DataFrame): DataFrame containing image paths.
        img_col (str): Column name containing image paths.
        img_width (int): Width of displayed images.
        img_height (int): Height of displayed images.

    Returns:
        IPython.display.HTML: HTML object for displaying images.

    """
    from IPython.display import HTML
    dfshow = df.copy()
    dfshow[img_col] = [f'<img src="{x}"  width="{img_width}" height="{img_height}" />'
                       for x in df[img_col]]
    return HTML(dfshow.to_html().replace('&lt;img', '<img').replace('/&gt;', '/>'))


class Colors:
    def __init__(self):
        self.hexs = ('FF3838', 'FF9D97', 'FF701F', 'FFB21D', 'CFD231', '48F90A', '92CC17', '3DDB86', '1A9334', '00D4BB',
                '2C99A8', '00C2FF', '344593', '6473FF', '0018EC', '8438FF', '520085', 'CB38FF', 'FF95C8', 'FF37C7')
        self.palette = [self.hex2rgb(f'#{c}') for c in self.hexs]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c
    
    def get_hex_color(self,i):
        return self.hexs[int(i)%self.n]

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i:1 + i + 2], 16) for i in (0, 2, 4))

colors = Colors()

class Annotator:
    def __init__(self, img, line_width=None, font_size=None):
        """
        Annotator class for adding bounding boxes, labels, and masks to images.

        Args:
            img (PIL.Image.Image or np.ndarray): Input image.
            line_width (int): Width of the bounding box lines.
            font_size (int): Font size for labels.
        """
        assert np.asarray(img).data.contiguous, 'Image not contiguous. Apply np.ascontiguousarray(im) to input images.'
        self.img = img if isinstance(img, Image.Image) else Image.fromarray(img)

        try:
            path = Path(__file__)
            font = str(path.parent/"assets/Arial.ttf")
            size = font_size or max(round(sum(self.img.size) / 2 * 0.035), 12)
            self.font = ImageFont.truetype(font, size)
        except Exception as err:
            print(err, file=sys.stderr)
            self.font = ImageFont.load_default()

        self.lw = line_width or max(round(sum(self.img.size) / 2 * 0.003), 2)

    def add_box_label(self, box, label='', color=(128, 128, 128), txt_color=(255, 255, 255)):
        """
        Add a bounding box with a label to the image.

        Args:
            box (tuple): Coordinates of the bounding box (x1, y1, x2, y2).
            label (str): Label text to display.
            color (tuple): RGB tuple representing the box color.
            txt_color (tuple): RGB tuple representing the label text color.

        Example:
            ```python
            annotator.add_box_label((x1, y1, x2, y2), label='example', color=(128, 128, 128), txt_color=(255, 255, 255))
            ```
        """
        self.draw = ImageDraw.Draw(self.img)

        self.draw.rectangle(box, width=self.lw, outline=color)  # box
        if label:
            
            _, _, w, h = self.font.getbbox(label)  # text width, height 
            outside = box[1] - h >= 0  # label fits outside box
            self.draw.rectangle(
                (box[0], box[1] - h if outside else box[1], box[0] + w + 1,
                 box[1] + 1 if outside else box[1] + h + 1),
                fill=color,
            )
            self.draw.text((box[0], box[1] - h if outside else box[1]), label, fill=txt_color, font=self.font)

    def add_mask(self, mask, color, alpha=0.5):
        """
        Add a mask to the image.

        Args:
            mask (np.ndarray): Binary mask array.
            color (tuple): RGB tuple representing the mask color.
            alpha (float): Opacity of the mask.

        Example:
            ```python
            annotator.add_mask(mask, color=(255, 0, 0), alpha=0.5)
            ```
        """
        img = np.asarray(self.img).copy()

        im1_shape = mask.shape
        im0_shape = img.shape

        gain = min(im1_shape[0] / im0_shape[0], im1_shape[1] / im0_shape[1])
        pad = (im1_shape[1] - im0_shape[1] * gain) / 2, (im1_shape[0] - im0_shape[0] * gain) / 2

        top, left = int(pad[1]), int(pad[0])  # y, x
        bottom, right = int(im1_shape[0] - pad[1]), int(im1_shape[1] - pad[0])
        mask = mask[top:bottom, left:right]

        mask_img = Image.fromarray((255 * mask).astype(np.uint8), mode='L')
        mask = np.array(mask_img.resize(self.img.size, resample=Image.BILINEAR)) >= 1
        img[mask] = img[mask] * (1 - alpha) + np.array(color) * alpha
        self.img = Image.fromarray(img)


def plot_detection(img, boxes, class_names, min_score=0.2):
    """
    Plot bounding boxes on an image for object detection.

    Args:
        img (PIL.Image.Image or np.ndarray): Input image.
        boxes (list): List of bounding boxes in the format [x1, y1, x2, y2].
        class_names (list): List of class names corresponding to the boxes.
        min_score (float): Minimum confidence score to display a box.

    Returns:
        PIL.Image.Image: Annotated image.
    """
    annotator = Annotator(img)
    for *box, conf, cls in reversed(boxes):
        if conf > min_score:
            label = f'{class_names[int(cls)]} {conf:.2f}'
            annotator.add_box_label(box, label, color=colors(cls))
    return annotator.img


def plot_instance_segmentation(img, boxes, masks, class_names, min_score=0.2):
    """
    Plot bounding boxes and masks on an image for instance segmentation.

    Args:
        img (PIL.Image.Image or np.ndarray): Input image.
        boxes (list): List of bounding boxes in the format [x1, y1, x2, y2].
        masks (list): List of masks corresponding to the boxes.
        class_names (list): List of class names corresponding to the boxes.
        min_score (float): Minimum confidence score to display a box.

    Returns:
        PIL.Image.Image: Annotated image.
    """
    annotator = Annotator(img)
    for (*box, conf, cls), mask in zip(reversed(boxes), reversed(masks)):
        if conf > min_score:
            label = f'{class_names[int(cls)]} {conf:.2f}'
            annotator.add_box_label(box, label, color=colors(cls))
            annotator.add_mask(mask.cpu().numpy(), colors(cls))
    return annotator.img


def vis_detection(image, prediction, class_names=None, min_score=0.2, figsize=(16, 16), linewidth=2, color='lawngreen'):
    """
    Visualize object detection results using Matplotlib.

    Args:
        image (PIL.Image.Image or np.ndarray): Input image.
        prediction (dict): Prediction results containing 'boxes', 'scores', and 'labels'.
        class_names (list): List of class names.
        min_score (float): Minimum confidence score to display a box.
        figsize (tuple): Size of the Matplotlib figure.
        linewidth (int): Width of the bounding box lines.
        color (str): Color of the bounding boxes.

    Example:
        ```python
        vis_detection(image, prediction, class_names, min_score=0.5)
        ```
    """
    boxes = np.array(prediction['boxes'].tolist())
    scores = np.array(prediction['scores'].tolist()) if 'scores' in prediction else [1.0 for _ in boxes]
    labels = np.array(prediction['labels'].tolist()) if 'labels' in prediction else None
    classes = [class_names[x] for x in labels] if (
        class_names is not None and labels is not None) else ['object'] * len(boxes)

    import matplotlib.pyplot as plt
    image = np.array(image, dtype=np.uint8)
    fig = plt.figure(figsize=figsize)
    plt.axis("off")
    plt.imshow(image)
    ax = plt.gca()

    for i, (box, name, score) in enumerate(zip(boxes, classes, scores)):
        if score >= min_score:
            text = "{}: {:.2f}".format(name, score)
            x1, y1, x2, y2 = box
            w, h = x2 - x1, y2 - y1
            patch = plt.Rectangle(
                [x1, y1], w, h, fill=False, edgecolor=color, linewidth=linewidth
            )
            ax.add_patch(patch)
            ax.text(
                x1,
                y1,
                text,
                bbox={"facecolor": color, "alpha": 0.8},
                clip_box=ax.clipbox,
                clip_on=True,
            )
    plt.show();


def joint_imgs_row(img1, img2):
    """
    Concatenate two images horizontally.

    Args:
        img1 (PIL.Image.Image or np.ndarray): First image.
        img2 (PIL.Image.Image or np.ndarray): Second image.

    Returns:
        PIL.Image.Image: Concatenated image.

    """
    size1 = img1.size
    size2 = img2.size
    joint = Image.new('RGB', (size1[0] + size2[0], max(size1[1], size2[1])))
    loc1, loc2 = (0, 0), (size1[0], 0)
    joint.paste(img1, loc1)
    joint.paste(img2, loc2)
    return joint


def joint_imgs_col(img1, img2):
    """
    Concatenate two images vertically.

    Args:
        img1 (PIL.Image.Image or np.ndarray): First image.
        img2 (PIL.Image.Image or np.ndarray): Second image.

    Returns:
        PIL.Image.Image: Concatenated image.

    """
    size1 = img1.size
    size2 = img2.size
    joint = Image.new('RGB', (max(size1[0], size2[0]), size1[1] + size2[1]))
    loc1, loc2 = (0, 0), (0, size1[1])
    joint.paste(img1, loc1)
    joint.paste(img2, loc2)
    return joint


def get_color_map_list(num_classes):
    """
    Get a color map for segmentation masks.

    Args:
        num_classes (int): Number of classes.
    """
    color_map = num_classes * [0, 0, 0]
    for i in range(0, num_classes):
        j = 0
        lab = i
        while lab:
            color_map[i * 3] |= (((lab >> 0) & 1) << (7 - j))
            color_map[i * 3 + 1] |= (((lab >> 1) & 1) << (7 - j))
            color_map[i * 3 + 2] |= (((lab >> 2) & 1) << (7 - j))
            j += 1
            lab >>= 3
    return color_map


def gray2pseudo(gray_img):
    """
    Convert a grayscale image to a pseudocolored image.

    Args:
        gray_img (Image): Grayscale image.

    Returns:
        Image: Pseudocolored image.
    """
    color_map = get_color_map_list(256)
    gray_p = gray_img.convert('P')
    gray_p.putpalette(color_map)
    return gray_p


def fig2img(fig):
    """
    Convert a Matplotlib figure to a PIL Image.

    Args:
        fig: Matplotlib figure.

    Returns:
        Image: PIL Image.
    """
    import io
    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = PIL.Image.open(buf)
    return img


def text2img(text):
    """
    Convert text to a PIL Image.

    Args:
        text (str): Text content.

    Returns:
        Image: PIL Image.
    """
    path = Path(__file__)
    simhei = path.parent / "assets/SimHei.ttf"
    lines = len(text.split("\n"))
    image = Image.new("RGB", (800, lines * 20), (255, 255, 255))
    draw = ImageDraw.Draw(image)
    font = ImageFont.truetype(str(simhei), 18)
    draw.text((0, 0), text, font=font, fill="#000000")
    return image


def img2tensor(image):
    """
    Convert a PIL Image to a PyTorch tensor.

    Args:
        image (Image): PIL Image.

    Returns:
        Tensor: PyTorch tensor.
    """
    from torchvision.transforms import ToTensor
    tensor = ToTensor()(np.array(image))
    return tensor



import math 
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from tqdm import tqdm
from omegaconf import DictConfig
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union
from ..base_model import BaseModel

class nn_Lambda(nn.Module):
    def __init__(self, fn, /, **kwargs) -> None:
        super().__init__()
        self._function = fn
        self._function_kwargs = kwargs
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Do the forward pass."""
        return self._function(x, **self._function_kwargs)


class CatEmbeddings(nn.Module):
    def __init__(
        self,
        categorical_embedding_dims,
        embedding_dropout = 0.0,
    ):
        super().__init__()
        self.categorical_embedding_dims = categorical_embedding_dims
        self.cat_embedding_layers = nn.ModuleList(
            [nn.Embedding(x, y) for x, y in categorical_embedding_dims])
        if embedding_dropout > 0:
            self.embd_dropout = nn.Dropout(embedding_dropout)
        else:
            self.embd_dropout = None

    def forward(self, x):
        # (B, N)
        x_cat = x
        assert x_cat.shape[1] == len(
            self.cat_embedding_layers
        ), "categorical_data must have same number of columns as categorical embedding layers"
        if x_cat.shape[1] > 0:
            embed = torch.cat(
                [
                    embedding_layer(x_cat[:, i])
                    for i, embedding_layer in enumerate(self.cat_embedding_layers)
                ],
                dim=1,
            )
        else:
            embed = torch.empty(0, 0)
        if self.embd_dropout is not None:
            embed = self.embd_dropout(embed)
        return embed

class OneHotEncoder(nn.Module):
    cardinalities: Tensor

    def __init__(self, cardinalities: List[int]) -> None:
        super().__init__()
        self.register_buffer('cardinalities', torch.tensor(cardinalities))

    def forward(self, x: Tensor) -> Tensor:
        encoded_columns = [
            F.one_hot(x[..., column], cardinality)
            for column, cardinality in zip(range(x.shape[-1]), self.cardinalities)
        ]
        return torch.cat(encoded_columns, -1)
        
class PeriodicEmbeddings(nn.Module):
    def __init__(
        self, n_features: int, n_frequencies: int, frequency_scale: float
    ) -> None:
        super().__init__()
        self.frequencies = nn.Parameter(
            torch.normal(0.0, frequency_scale, (n_features, n_frequencies))
        )
    def forward(self, x: Tensor) -> Tensor:
        assert x.ndim == 2
        x = 2 * torch.pi * self.frequencies[None] * x[..., None]
        x = torch.cat([torch.cos(x), torch.sin(x)], -1)
        return x


class PLREmbeddings(nn.Sequential):
    def __init__(
        self,
        n_features: int,
        n_frequencies: int,
        frequency_scale: float,
        d_embedding: int,
    ) -> None:
        super().__init__(
            PeriodicEmbeddings(n_features, n_frequencies, frequency_scale),
            nn.Linear(2 * n_frequencies, d_embedding),
            nn.ReLU(),
        )

class DeepKNNModel(nn.Module):
    def __init__(self, config, **kwargs):
        super().__init__()
        self.hparams = config
        
        d_numerical = self.hparams.continuous_dim
        d_main = self.hparams.hidden_size
        dropout0 = self.hparams.block_dropout0
        categories = self.hparams.categorical_cardinality
        n_classes = self.hparams.output_dim
        self.d_block = int(2.0*d_main)
        
        self.cat_embeddings = CatEmbeddings(
            categorical_embedding_dims = self.hparams.embedding_dims
        ) if categories else None
        
        self.num_embeddings = PLREmbeddings(
            n_features = d_numerical,
            n_frequencies = self.hparams.plr_frequencies_num,
            frequency_scale = self.hparams.plr_frequency_scale,
            d_embedding = self.hparams.plr_embed_dim
        ) if d_numerical>0 else None

        d_in = d_numerical * self.hparams.plr_embed_dim + self.hparams.embedded_cat_dim
        
        self.linear = nn.Linear(d_in, d_main)
        self.blocks0 = nn.ModuleList(
            [self.make_block(i > 0) 
             for i in range(self.hparams.encoder_block_num)]
        )

        #R
        self.normalization = nn.LayerNorm(d_main) if self.hparams.encoder_block_num>0 else None 
        self.label_encoder = (
            nn.Linear(1, d_main)
            if n_classes==1
            else nn.Sequential(
                nn.Embedding(n_classes, d_main), nn_Lambda(lambda x: x.squeeze(-2))
            )
        )

        self.K = nn.Linear(d_main, d_main)
        self.T = nn.Sequential(
            nn.Linear(d_main, self.d_block),
            nn.ReLU(),
            nn.Dropout(dropout0),
            nn.Linear(self.d_block, d_main, bias=False),
        )
        self.dropout = nn.Dropout(self.hparams.context_dropout)

        #P
        self.blocks1 = nn.ModuleList(
            [self.make_block(True) for _ in range(self.hparams.predictor_block_num)]
        )
        
        self.major_head = nn.Sequential(
            nn.LayerNorm(d_main),
            nn.ReLU(),
            nn.Linear(d_main, n_classes),
        )
        
        self.aux_head = nn.Sequential(
            nn.LayerNorm(d_main),
            nn.ReLU(),
            nn.Linear(d_main, n_classes),
        )

        self.loss = getattr(nn, self.hparams.loss)()
        
        self.search_index = None
        self.pretrain_encoder = False
        self.reset_parameters()
        
    def make_block(self, prenorm) -> nn.Sequential:
        return nn.Sequential(
                    *([nn.LayerNorm(self.hparams.hidden_size)] if prenorm else []),
                    nn.Linear(self.hparams.hidden_size, self.d_block),
                    nn.ReLU(),
                    nn.Dropout(self.hparams.block_dropout0),
                    nn.Linear(self.d_block, self.hparams.hidden_size),
                    nn.Dropout(self.hparams.block_dropout1),
        )

    def reset_parameters(self):
        if isinstance(self.label_encoder, nn.Linear):
            bound = 1 / math.sqrt(2.0)
            nn.init.uniform_(self.label_encoder.weight, -bound, bound)  
            nn.init.uniform_(self.label_encoder.bias, -bound, bound) 
            if self.hparams.output_dim>1:
                assert isinstance(self.label_encoder[0], nn.Embedding)
                nn.init.uniform_(self.label_encoder[0].weight, -1.0, 1.0)  

    def encode(self, inputs):
        x_cat,x_num = inputs["categorical"], inputs["continuous"]
        x = []
        if self.num_embeddings is not None:
            x.append(
                x_num
                if self.num_embeddings is None
                else self.num_embeddings(x_num).flatten(1)
            )
            
        if self.cat_embeddings is not None:
            x.append(self.cat_embeddings(x_cat))
        assert x
        
        x = torch.cat(x, dim=1)
        x = self.linear(x)
        for block in self.blocks0:
            x = x + block(x)
        k = self.K(x if self.normalization is None else self.normalization(x))
        return x, k

    @torch.no_grad()
    def update_search_index(self, dl_candidates):
        candidate_k_list = []
        candidate_y_list = []
        candidate_id_list = []
        n = dl_candidates.size if hasattr(dl_candidates, 'size') else len(dl_candidates)
        
        loop = tqdm(enumerate(dl_candidates, start=1),total=n,
                    disable=True,ncols=100)
        
        for i,batch in loop:
            candidate_k_list.append(self.encode(batch)[1])
            candidate_y_list.append(batch['target'])
            candidate_id_list.append(batch['id'])

        self.candidate_k = torch.cat(candidate_k_list).cpu()
        self.candidate_y = torch.cat(candidate_y_list).cpu()
        self.candidate_id = torch.cat(candidate_id_list).cpu()
        self.dl_candidates = dl_candidates
        
        num_candidate, d_main = self.candidate_k.shape
        device = self.candidate_k.device
        
        if self.search_index is None:
            import faiss
            import faiss.contrib.torch_utils
            self.search_index = faiss.IndexHNSWFlat(d_main,16) 
                
        self.search_index.reset()
        self.search_index.train(self.candidate_k)
        self.search_index.add(self.candidate_k)
        
    @torch.no_grad()
    def find_context(self, inputs):
        x, k = self.encode(inputs)
        input_id = inputs['id']
        
        context_size = self.hparams.context_size + (1 if self.training else 0) 
        distances, context_idx = self.search_index.search(k.cpu(), context_size)
        context_id = self.candidate_id[context_idx].to(input_id.device)
        distances = distances.to(input_id.device)

        #delete sample with the same id in inputs, in case of a lable leakage.
        if self.training:
            context_id = torch.where(context_id!=input_id[:,None],context_id,-1)
            distances[context_id<0] = torch.inf
            context_id = context_id.gather(-1, distances.argsort()[:, :-1])

        ds_candidates = self.dl_candidates.dataset
        batch_ids = context_id.flatten().cpu()
        contexts = ds_candidates.get_batch(batch_ids)
        contexts = {key:value.to(input_id.device) for key,value in contexts.items()}
        return contexts 

    
    def forward(
        self,
        inputs,
        contexts = None
    ):
        x, k = self.encode(inputs)

        #pretrain_encoder
        if self.pretrain_encoder:
            aux_logits = self.aux_head(x)
            output = {'logits': aux_logits}
            return output 
        
        y = inputs.get('target',None)
        if contexts is None:
            contexts = self.find_context(inputs)

        batch_size, d_main = k.shape 
        context_size = self.hparams.context_size 
        
        context_k = self.encode(contexts)[1].reshape(batch_size,context_size,d_main)
        context_y = contexts['target'].reshape(batch_size,context_size)

        similarities = (
            -k.square().sum(-1, keepdim=True)
            + (2 * (k[..., None, :] @ context_k.transpose(-1, -2))).squeeze(-2)
            - context_k.square().sum(-1)
        )
        probs = F.softmax(similarities, dim=-1)
        probs = self.dropout(probs)

        context_y_emb = self.label_encoder(context_y[..., None])
        values = context_y_emb + self.T(k[:, None] - context_k)
        context_x = (probs[:, None] @ values).squeeze(1)
        x = x + context_x

        # >>>
        for block in self.blocks1:
            x = x + block(x)
        logits = self.major_head(x)
        output = {'logits': logits}
        return output 
        
    def compute_loss(self,output, y):
        y_hat = output["logits"]
        computed_loss = self.loss(y_hat.squeeze(), y.squeeze())
        return computed_loss 




import torch 
from torch.utils.data import Dataset 
import numpy as np 
import pandas as pd 
from typing import Dict, Iterable, List, Optional, Tuple, Union

class TabularDataset(Dataset):
    def __init__(
        self,
        data: pd.DataFrame,
        task: str,
        continuous_cols: List[str] = None,
        categorical_cols: List[str] = None,
        target: List[str] = None,
        start_id: int = 0,
    ):
        """Dataset to Load Tabular Data.

        Args:
            data (pd.DataFrame): Pandas DataFrame to load during training
            task (str):
                It is a regression/binary/multilass  task. If multilass, it returns a LongTensor as target
            continuous_cols (List[str], optional): A list of names of continuous columns. Defaults to None.
            categorical_cols (List[str], optional): A list of names of categorical columns.
                These columns must be ordinal encoded beforehand. Defaults to None.
            target (List[str], optional): A list of strings with target column name(s). Defaults to None.

        """

        self.task = task
        self.n = data.shape[0]
        self.target = target
        self.start_id = start_id 
        self.ids = np.arange(start_id,start_id+self.n)
        if target:
            self.y = data[target].astype(np.float32).values
            if isinstance(target, str):
                self.y = self.y.reshape(-1, 1)  
        else:
            self.y = np.zeros((self.n, 1))  

        if task in ("multiclass","classification"):
            self.y = self.y.astype(np.int64)
   
        self.categorical_cols = categorical_cols if categorical_cols else []
        self.continuous_cols = continuous_cols if continuous_cols else []

        if self.continuous_cols:
            self.continuous_X = data[self.continuous_cols].astype(np.float32).values

        if self.categorical_cols:
            self.categorical_X = data[categorical_cols]
            self.categorical_X = self.categorical_X.astype(np.int64).values

    @property
    def data(self):
        """Returns the data as a pandas dataframe."""
        if self.continuous_cols and self.categorical_cols:
            data = pd.DataFrame(
                np.concatenate([self.categorical_X, self.continuous_X], axis=1),
                columns=self.categorical_cols + self.continuous_cols,
            )
        elif self.continuous_cols:
            data = pd.DataFrame(self.continuous_X, columns=self.continuous_cols)
        elif self.categorical_cols:
            data = pd.DataFrame(self.categorical_X, columns=self.categorical_cols)
        else:
            data = pd.DataFrame()
        for i, t in enumerate(self.target):
            data[t] = self.y[:, i]
        return data

    def __len__(self):
        """Denotes the total number of samples."""
        return self.n

    def __getitem__(self, idx):
        """Generates one sample of data."""
        return {
            "target": torch.tensor(self.y[idx]),
            "continuous": (torch.tensor(self.continuous_X[idx]) if self.continuous_cols else torch.Tensor()),
            "categorical": (torch.tensor(self.categorical_X[idx]) if self.categorical_cols else torch.Tensor()),
            "id": torch.tensor(self.ids[idx],dtype=torch.long),
        }
    def get_batch(self, batch_ids):
        batch_idxes = batch_ids - self.start_id
        batch = {
            "target": torch.tensor(self.y[batch_idxes]),
            "continuous": (torch.tensor(self.continuous_X[batch_idxes]) if self.continuous_cols else torch.Tensor()),
            "categorical": (torch.tensor(self.categorical_X[batch_idxes]) if self.categorical_cols else torch.Tensor()),
            "id": torch.tensor(self.ids[batch_idxes],dtype=torch.long),
        }
        return batch 
        




class ChatGPT(object):
    def __init__(self,
                 api_key = None,
                 max_chat_rounds=3,
                 proxy = False,
                 model="gpt-3.5-turbo",
                 temperature=0,
                 system_content='You are friendly chatbot.'
                ):
        
        import json
        from pathlib import Path 
        json_file = Path(__file__).parent/'openai.json'
        if api_key is not None:
            with open(json_file,'w') as fp:
                json.dump({'api_key':api_key},fp)
        else:
            with open(json_file,'r') as fp:
                api_key = json.load(fp)['api_key']
        
        self.api_key = api_key 
        self.model = model
        self.max_chat_rounds = max_chat_rounds
        self.temperature = temperature
        self.messages =  [{'role':'system', 'content':system_content}]
        
        if  proxy==False:
            import openai
            openai.api_key  = api_key
            test_prompt = [{'role':'user','content':'hello'}]
            reply = self.get_openai_completion(self.messages+test_prompt)
            self.get_completion_from_messages = self.get_openai_completion
            print(reply)
            
        else:
            url = "https://closeai.deno.dev/v1/chat/completions"
            #url = "https://chatai.1rmb.tk/api/v1/chat/completions"
            #url =  "https://api.lixining.com/v1/chat/completions"  
            self.proxy = url if not isinstance(proxy,str) else proxy 
            test_prompt = [{'role':'user','content':'你好'}]
            reply = self.get_proxy_completion(self.messages+test_prompt)
            self.get_completion_from_messages = self.get_proxy_completion
            print(reply)
        try:
            self.register_magic() 
            print('register magic %%chat sucessed ...')
        except Exception as err:
            print('register magic %%chat failed ...')
            print(err)
            
    def get_openai_completion(self,messages):
        import openai 
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=messages,
            temperature=self.temperature, # this is the degree of randomness of the model's output
        )
        return response.choices[0].message["content"]

    def get_proxy_completion(self,messages):  
        
        import requests 
        headers = {
          'Authorization': f'Bearer {self.api_key}',
          'Content-Type': 'application/json'
        }

        payload = {
          "model": self.model,
          "messages": messages
        }
        
        try:
            response = requests.post(self.proxy, headers=headers, json=payload)
            response.raise_for_status() 
            data = response.json()
            result = (data["choices"][0]["message"]["content"])
            return result
        except Exception as err:
            print(f"error: {err}")
            return err
    
    def __call__(self,prompt):
        if len(self.messages)>=2*self.max_chat_rounds+1:
            self.messages = [self.messages[0]]+self.messages[3:]
        self.messages.append({'role':'user','content':prompt})
        reply = self.get_completion_from_messages(messages = self.messages)
        self.messages.append({'role':'assistant','content':reply})
        return reply  
    
    def register_magic(self):
        import IPython
        from IPython.core.magic import (Magics, magics_class, line_magic,
                                        cell_magic, line_cell_magic)
        @magics_class
        class ChatGPTMagics(Magics):
            def __init__(self,shell, pipe):
                super().__init__(shell)
                self.pipe = pipe

            @line_cell_magic
            def chat(self, line, cell=None):
                "Magic that works both as %chat and as %%chat"
                if cell is None:
                    return self.pipe(line)
                else:
                    print(self.pipe(cell))
                    
        ipython = IPython.get_ipython()
        magic = ChatGPTMagics(ipython,self)
        ipython.register_magics(magic)


from dataclasses import dataclass, field
from typing import Optional,List 

from torchkeras.tabular.config import ModelConfig

@dataclass
class DeepCrossConfig(ModelConfig):
    input_embed_max: int = field(
        default=64,
        metadata={"help": 
        "The max embedding dimension for the input categorical features. Defaults to 64."
        },
    )

    cross_type: str = field(
        default = 'matrix',
        metadata={
            "help": "cross type( vector for DCN-vector, matrix for DCN-matrix, mix for DCN-mix)",
            "choices": ['vector','matrix','mix'],
        }
    )
    
    cross_order: int = field(
        default = 2,
        metadata={
            "help": "cross order level",
            "choices": [2, 3,1],
        }
    )

    low_rank: int = field(
        default = 32,
        metadata={"help": "low rank for MOE, only used for DCN-mix"}
    )
    

    experts_num: int = field(
        default = 4,
        metadata={"help": "experts number for MOE, only used for DCN-mix"}
    )

    mlp_layers: str = field(
        default = "128-64-32",
        metadata={"help": "Hyphen-separated number of layers and units in the MLP. Defaults to 128-64-32"},
    )

    mlp_dropout: float = field(
        default = 0.25,
        metadata={"help": "The dropout ratio for  MLP. Defaults to 0.25"},
    )
    

    stacked: bool = field(
        default = False,
        metadata={"help": "stack or parallel the deep part and cross part"},
    )

    embedding_initialization: Optional[str] = field(
        default="kaiming_uniform",
        metadata={
            "help": "Initialization scheme for the embedding layers. Defaults to `kaiming`",
            "choices": ["kaiming_uniform", "kaiming_normal"],
        }
    )
    
    _module_src: str = field(default="models.deepcross")
    _model_name: str = field(default="DeepCrossModel")
    _backbone_name: str = field(default="DeepCrossBackbone")
    _config_name: str = field(default="DeepCrossConfig")
    
    


'''
reference from https://github.com/fastai/fastprogress/
'''
import time,sys
from IPython.display import clear_output, display, HTML
from tqdm.utils import _term_move_up

"""
@author : lyhue1991
@description : pbar
"""

move_up = _term_move_up()

def is_jupyter():
    """
    Check if the code is running in a Jupyter notebook environment.

    Returns:
        bool: True if running in Jupyter notebook, False otherwise.
    """
    import contextlib
    with contextlib.suppress(Exception):
        from IPython import get_ipython
        return get_ipython() is not None
    return False


def format_time(t):
    """
    Format time in seconds to (h):mm:ss.

    Args:
        t (int): Time in seconds.

    Returns:
        str: Formatted time string.
    """
    t = int(t)
    h, m, s = t // 3600, (t // 60) % 60, t % 60
    if h != 0:
        return f'{h}:{m:02d}:{s:02d}'
    else:
        return f'{m:02d}:{s:02d}'


def format_number(x):
    """
    Format a number in scientific notation if too small or too large.

    Args:
        x (float): Number to be formatted.

    Returns:
        str: Formatted number string.
    """
    if abs(x) < 1e-4 or abs(x) >= 1e5:
        return "{:.4e}".format(x)
    else:
        return "{:.4f}".format(x)


# HTML styles for progress bar
html_progress_bar_styles = """
<style>
    /* background: */
    progress::-webkit-progress-bar {background-color: #CDCDCD; width: 100%;}
    progress {background-color: #CDCDCD;}

    /* value: */
    progress::-webkit-progress-value {background-color: #00BFFF  !important;}
    progress::-moz-progress-bar {background-color: #00BFFF  !important;}
    progress {color: #00BFFF ;}

    /* optional */
    .progress-bar-interrupted, .progress-bar-interrupted::-webkit-progress-bar {
        background: #000000;
    }
</style>
"""

def html_progress_bar(value, total, label='', postfix='', interrupted=False):
    """
    Generate HTML code for a progress bar.

    Args:
        value (float): Current progress value.
        total (float): Total progress value.
        label (str): Additional label for the progress bar.
        postfix (str): Additional information to display after the progress bar.
        interrupted (bool): Whether the progress bar is interrupted.

    Returns:
        str: HTML code for the progress bar.
    """
    bar_style = 'progress-bar-interrupted' if interrupted else ''
    val = '' if total is None else f"value='{value}'"
    return f"""
    <div>
      <progress {val} class='{bar_style}' max='{total}' style='width:300px; height:20px; vertical-align: middle;'></progress>
      {label}
      <br>
      {postfix}
    </div>
    """


def text_progress_bar(value, total, label='', postfix='', interrupted=False):
    """
    Generate a text-based progress bar.

    Args:
        value (float): Current progress value.
        total (float): Total progress value.
        label (str): Additional label for the progress bar.
        postfix (str): Additional information to display after the progress bar.
        interrupted (bool): Whether the progress bar is interrupted.

    Returns:
        str: Text-based progress bar string.
    """
    bar_style = "🟦" if interrupted else "⬜️"  # Google style: "🟥", "⬜️" "○", "*"'🟦''⬛️'

    # Calculate the percentage of completion
    percentage = round(value / total * 20)

    # Define the finished and unfinished parts of the progress bar
    finished = "🟩" * percentage  # Google style: "🟩", "●"
    unfinished = bar_style * (20 - percentage)

    # Construct the progress bar string
    bar = "\r{}{} {}".format(finished, unfinished, label) + " " * 20 + "\t" * 50 + f"{postfix}" + " " * 20 + "\t" * 50
    return bar

class ProgressBar:
    update_every, first_its, lt = 0.2, 5, '<'

    def __init__(self, gen, total=None, comment='', comment_tail=''):
        """
        Initialize the ProgressBar instance.

        Args:
            gen: Iterable to track progress.
            total: Total number of iterations (None for unknown).
            comment: Initial comment to display.
            comment_tail: Additional tail for the comment.
        """
        # Initialize instance variables
        self.gen, self.comment, self.comment_tail = gen, comment, comment_tail
        self.postfix = ''
        self.total = None if total == 'noinfer' else len(gen) if total is None else total
        self.last_v = None
        self.display = True
        self.in_jupyter = is_jupyter()
        self.update(0)

    def update(self, val):
        """
        Update the progress bar.

        Args:
            val: Current value of the progress.
        """
        # Initialization on the first iteration
        if self.last_v is None:
            self.on_iter_begin()
            self.last_v = 0

        # Initial update or update triggered by specific conditions
        if val == 0:
            self.start_t = self.last_t = time.time()
            self.pred_t, self.last_v, self.wait_for = None, 0, 1
            self.update_bar(0)
        elif val <= self.first_its or val >= self.last_v + self.wait_for or (self.total and val >= self.total):
            cur_t = time.time()
            avg_t = (cur_t - self.start_t) / val
            self.wait_for = max(int(self.update_every / (avg_t + 1e-8)), 1)
            self.pred_t = None if self.total is None else avg_t * self.total
            self.last_v, self.last_t = val, cur_t

            # Check if the last iteration is reached
            if self.total is not None and val >= self.total:
                self.on_iter_end()
                self.last_v = self.total
            else:
                self.update_bar(val)

    def on_iter_begin(self):
        """
        Initialize progress bar at the beginning of an iteration.
        """
        if self.in_jupyter:
            # Display progress bar in Jupyter environment
            self.html_code = '\n'.join([html_progress_bar(0, self.total, ""), ""])
            display(HTML(html_progress_bar_styles))
            self.out = display(HTML(self.html_code), display_id=True)
        else:
            print('\n')

    def on_iter_end(self):
        """
        Handle the end of an iteration.
        """
        total_time = format_time(time.time() - self.start_t)
        self.comment = f'100% [{self.total}/{self.total}] [{total_time}]'
        self.on_update(self.total, self.comment, self.postfix, False, 1)
        self.display = False
        if not self.in_jupyter:
            print('\n')

    def on_update(self, val, comment='', postfix='', interrupted=False, up=1):
        """
        Update the progress bar.

        Args:
            val: Current value of the progress.
            comment: Comment to display.
            postfix: Additional information to display.
            interrupted: Whether the process was interrupted.
            up: Number of lines to move up.
        """
        if not self.display:
            return
        if self.in_jupyter:
            # Update HTML progress bar in Jupyter environment
            self.progress = html_progress_bar(val, self.total, comment, postfix, interrupted)
            self.out.update(HTML(self.progress))
        else:
            if self.comment_tail:
                comment = comment + f' [{self.comment_tail}]'
            progress = text_progress_bar(val, self.total, comment, postfix, interrupted)
            print(move_up * up + progress, end='')

    def on_interrupt(self, msg='interrupted'):
        """
        Handle the case of interruption.

        Args:
            msg: Message to display for the interruption.
        """
        comment = self.comment + f' [{msg}]' if msg else self.comment
        self.on_update(self.last_v, comment, self.postfix, interrupted=True, up=1)
        if not self.in_jupyter:
            print('\n')

    def __iter__(self):
        """
        Iterate over the provided generator.

        Yields:
            Output from the generator.
        """
        if self.total != 0:
            self.update(0)
        try:
            for i, o in enumerate(self.gen):
                if self.total and i >= self.total:
                    break
                yield o
                self.update(i + 1)
            if self.total is None and self.last_v is not None:
                self.total = i + 1
                self.update(self.total)
        except Exception as e:
            self.on_interrupt()
            raise e

    def update_bar(self, val):
        """
        Update the progress bar details.

        Args:
            val: Current value of the progress.
        """
        if self.total == 0:
            return self.on_update(0, '100% [0/0]')
        if val == 0:
            self.comment = f'0% [0/{self.total}]'
            return self.on_update(0, self.comment)
        pct = '' if self.total is None else f'{100 * val / self.total:.2f}%'
        tot = '?' if self.total is None else str(self.total)
        elapsed_t = self.last_t - self.start_t
        remaining_t = '?' if self.pred_t is None else format_time(self.pred_t - elapsed_t)
        elapsed_t = format_time(elapsed_t)
        self.comment = f'{pct} [{val}/{tot}] [{elapsed_t}{self.lt}{remaining_t}]'
        self.on_update(val, self.comment, self.postfix)

    def set_postfix(self, **kwargs):
        """
        Set the postfix information for the progress bar.

        Args:
            **kwargs: Additional information to display.
        """
        if not self.display:
            return
        postfix = ''
        if 'i' in kwargs and 'n' in kwargs:
            from tqdm.std import Bar
            i, n = kwargs['i'], kwargs['n']
            kwargs.pop('i')
            kwargs.pop('n')
            ratio = i / n
            postfix += format(Bar(ratio, default_len=20))
            postfix += f'{100 * i/n:.2f}%'
            postfix += f' [{i}/{n}]'
        if kwargs:
            postfix += ' ['
            for i, (key, value) in enumerate(kwargs.items()):
                if isinstance(value, float):
                    postfix = postfix + f'{key}={format_number(value)}, '
                else:
                    postfix = postfix + f'{key}={value}, '
            postfix = postfix[:-2] + ']'
        self.postfix = postfix
        self.on_update(self.last_v, self.comment, self.postfix)



from .category_embedding_model import CategoryEmbeddingBackbone, CategoryEmbeddingModel
from .config import CategoryEmbeddingModelConfig

__all__ = ["CategoryEmbeddingModel", "CategoryEmbeddingModelConfig", "CategoryEmbeddingBackbone"]



import torch.nn as nn
import torch

class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_channel, out_channel, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=in_channel, out_channels=out_channel,
                               kernel_size=1, stride=1, bias=False)  # squeeze channels
        self.bn1 = nn.BatchNorm2d(out_channel)
        # -----------------------------------------
        self.conv2 = nn.Conv2d(in_channels=out_channel, out_channels=out_channel,
                               kernel_size=3, stride=stride, bias=False, padding=1)
        self.bn2 = nn.BatchNorm2d(out_channel)
        # -----------------------------------------
        self.conv3 = nn.Conv2d(in_channels=out_channel, out_channels=out_channel*self.expansion,
                               kernel_size=1, stride=1, bias=False)  # unsqueeze channels
        self.bn3 = nn.BatchNorm2d(out_channel*self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

    def forward(self, x):
        identity = x
        if self.downsample is not None:
            identity = self.downsample(x)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, blocks_num, num_classes=1000, include_top=True):
        super(ResNet, self).__init__()
        self.include_top = include_top
        self.in_channel = 64

        self.conv1 = nn.Conv2d(3, self.in_channel, kernel_size=7, stride=2,
                               padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(self.in_channel)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, blocks_num[0])
        self.layer2 = self._make_layer(block, 128, blocks_num[1], stride=2)
        self.layer3 = self._make_layer(block, 256, blocks_num[2], stride=2)
        self.layer4 = self._make_layer(block, 512, blocks_num[3], stride=2)
        if self.include_top:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))  # output size = (1, 1)
            self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')

    def _make_layer(self, block, channel, block_num, stride=1):
        downsample = None
        if stride != 1 or self.in_channel != channel * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.in_channel, channel * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(channel * block.expansion))

        layers = []
        layers.append(block(self.in_channel, channel, downsample=downsample, stride=stride))
        self.in_channel = channel * block.expansion

        for _ in range(1, block_num):
            layers.append(block(self.in_channel, channel))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        if self.include_top:
            x = self.avgpool(x)
            x = torch.flatten(x, 1)
            x = self.fc(x)

        return x

class ResNet50(ResNet):
    def __init__(self,num_classes=1000, include_top=True):
        super().__init__(Bottleneck, [3, 4, 6, 3], 
                 num_classes=num_classes, include_top=include_top)
    
def resnet50(num_classes=1000, include_top=True):
    net = ResNet(Bottleneck, [3, 4, 6, 3], 
                 num_classes=num_classes, include_top=include_top)
    return net




import pandas as pd
import numpy as np 
import matplotlib.pyplot as plt
from transformers import TrainerCallback
from torchkeras.pbar import is_jupyter

class VLogCallback(TrainerCallback):
    def __init__(self, figsize=(6,4), update_freq=1, save_path='history.png'):
        self.figsize = figsize
        self.update_freq = update_freq
        self.save_path = save_path
        self.in_jupyter = is_jupyter()

    def on_train_begin(self, args, state, control, **kwargs):
        metric = args.metric_for_best_model
        self.greater_is_better = args.greater_is_better 
        self.prefix = 'val_' if metric.startswith('val_') else 'eval_'
        self.metric =  metric
        
        dfhistory = pd.DataFrame()
        x_bounds = [0, args.logging_steps*10]
        self.update_graph(dfhistory, self.metric.replace(self.prefix,''), 
                             x_bounds = x_bounds, 
                             figsize = self.figsize)
        
    def on_evaluate(self, args, state, control, metrics, **kwargs):
        dfhistory = self.get_history(state)
        n = dfhistory['step'].max()
        if n%self.update_freq==0:
            x_bounds = [dfhistory['step'].min(), 
                        10*args.logging_steps+(n//(args.logging_steps*10))*args.logging_steps*10]
            self.update_graph(dfhistory, self.metric.replace(self.prefix,''), 
                              x_bounds = x_bounds, figsize = self.figsize)
            
    def on_train_end(self, args, state, control, **kwargs):
        dfhistory = self.get_history(state)
        self.update_graph(dfhistory, self.metric.replace(self.prefix,''), 
                             figsize = self.figsize)
        
    def get_history(self,state):
        log_history = state.log_history  
        train_history = [x for x in log_history if 'loss' in x.keys()]
        eval_history = [x for x in log_history if 'eval_loss'  in x.keys()]

        dfhistory_train = pd.DataFrame(train_history)
        dfhistory_eval = pd.DataFrame(eval_history)  
        dfhistory = dfhistory_train.merge(dfhistory_eval,on=['step','epoch'])
        if self.prefix=='val_':
            dfhistory.columns = [x.replace('eval_','val_') for x in dfhistory.columns]
        return dfhistory
    
    def get_best_score(self, dfhistory):
        arr_scores = dfhistory[self.metric]
        best_score = np.max(arr_scores) if self.greater_is_better==True else np.min(arr_scores)
        best_step = dfhistory.loc[arr_scores==best_score,'step'].tolist()[0]
        return (best_step, best_score)

    def update_graph(self, dfhistory, metric, x_bounds=None, 
                     y_bounds=None, figsize=(6,4)):
        
        from IPython.display import display
        if not hasattr(self, 'graph_fig'):
            self.graph_fig, self.graph_ax = plt.subplots(1, figsize=figsize)
            self.graph_out = display(self.graph_ax.figure, display_id=True)

        self.graph_ax.clear()
        steps = dfhistory['step'] if 'step' in dfhistory.columns else []

        m1 = metric
        if  m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            self.graph_ax.plot(steps,train_metrics,'bo--',label= m1,clip_on=False)

        m2 = self.prefix+metric
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.graph_ax.plot(steps,val_metrics,'co-',label =m2,clip_on=False)

        self.graph_ax.set_xlabel("step")
        self.graph_ax.set_ylabel(metric)  

        if m1 in dfhistory.columns or m2 in dfhistory.columns or metric in dfhistory.columns:
            self.graph_ax.legend(loc='best')
            
        if len(steps)>0:
            best_step, best_score = self.get_best_score(dfhistory)
            self.graph_ax.plot(best_step,best_score,'r*',markersize=15,clip_on=False)
            title = f'best {self.metric} = {best_score:.4f} (@step {best_step})'
            self.graph_ax.set_title(title)
        else:
            title = f'best {self.metric} = ?'
            self.graph_ax.set_title(title)
            
        if x_bounds is not None: self.graph_ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.graph_ax.set_ylim(*y_bounds)
        if self.in_jupyter:
            self.graph_out.update(self.graph_ax.figure)
        self.graph_fig.savefig(self.save_path)
        plt.close();
        
if __name__=='__main__':
    import os
    import numpy as np 
    import pandas as pd 
    import torch 
    import datasets 
    from transformers import AutoTokenizer,DataCollatorWithPadding
    from transformers import AutoModelForSequenceClassification 
    from transformers import TrainingArguments,Trainer 
    from transformers import EarlyStoppingCallback

    from tqdm import tqdm 
    from transformers import AdamW, get_scheduler
    from torchkeras.tools.transformers import VLogCallback 
    
    #================================================================================
    # 1,prepare data
    
    data_url = 'https://github.com/lyhue1991/torchkeras/releases/download/v3.7.2/waimai_10k.csv'
    data_file = 'waimai_10k.csv'

    if not os.path.exists(data_file):
        torch.hub.download_url_to_file(data_url,data_file)


    df = pd.read_csv("waimai_10k.csv")
    ds = datasets.Dataset.from_pandas(df)
    ds = ds.shuffle(42) 
    ds = ds.rename_columns({"review":"text","label":"labels"})

    tokenizer = AutoTokenizer.from_pretrained('bert-base-chinese') 

    ds_encoded = ds.map(lambda example:tokenizer(example["text"]),
                        remove_columns = ["text"],
                        batched=True)

    #train,val,test split
    ds_train_val,ds_test = ds_encoded.train_test_split(test_size=0.2).values()
    ds_train,ds_val = ds_train_val.train_test_split(test_size=0.2).values() 

    data_collator = DataCollatorWithPadding(tokenizer=tokenizer)

    dl_train = torch.utils.data.DataLoader(ds_train, batch_size=16, collate_fn = data_collator)
    dl_val = torch.utils.data.DataLoader(ds_val, batch_size=16,  collate_fn = data_collator)
    dl_test = torch.utils.data.DataLoader(ds_test, batch_size=16,  collate_fn = data_collator)

    for batch in dl_train:
        break
    print({k: v.shape for k, v in batch.items()})
    
    #================================================================================
    #2，define model
    
    model = AutoModelForSequenceClassification.from_pretrained(
        'bert-base-chinese',num_labels=2)


    #================================================================================
    #3，train model
    
    def compute_metrics(eval_preds):
        logits, labels = eval_preds
        preds = np.argmax(logits, axis=-1)
        accuracy = np.sum(preds==labels)/len(labels)
        precision = np.sum((preds==1)&(labels==1))/np.sum(preds==1)
        recall = np.sum((preds==1)&(labels==1))/np.sum(labels==1)
        f1  = 2*recall*precision/(recall+precision)
        return {"accuracy":accuracy,"precision":precision,"recall":recall,'f1':f1}

    training_args = TrainingArguments(
        output_dir = "bert_waimai",
        num_train_epochs = 2,
        logging_steps = 20,
        gradient_accumulation_steps = 10,
        evaluation_strategy="steps", #epoch

        metric_for_best_model='eval_f1',
        greater_is_better=True,

        report_to='none',
        load_best_model_at_end=True
    )

    callbacks = [EarlyStoppingCallback(early_stopping_patience=10),
                 VLogCallback()]

    trainer = Trainer(
        model,
        training_args,
        train_dataset=ds_train,
        eval_dataset=ds_val,
        compute_metrics=compute_metrics,
        callbacks = callbacks,
        data_collator=data_collator,
        tokenizer=tokenizer,
    )
    trainer.train() 


    #================================================================================
    #4，evaluate model
    
    trainer.evaluate(ds_val)


    #================================================================================
    #5，use model
    
    from transformers import pipeline
    model.config.id2label = {0:"差评",1:"好评"}
    classifier = pipeline(task="text-classification",tokenizer = tokenizer,model=model.cpu())
    classifier("挺好吃的哦")

    #================================================================================
    #6，save model
    
    model.save_pretrained("waimai_10k_bert")
    tokenizer.save_pretrained("waimai_10k_bert")

    classifier = pipeline("text-classification",model="waimai_10k_bert")
    classifier(["味道还不错，下次再来","我去，吃了我吐了三天"])



from functools import partial

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..common.layers.activations import entmax15
from ..common.layers.batch_norm import GBN


def initialize_glu(module, input_dim, output_dim):
    gain_value = np.sqrt((input_dim + output_dim) / np.sqrt(input_dim))
    torch.nn.init.xavier_normal_(module.weight, gain=gain_value)
    return


class LearnableLocality(nn.Module):
    def __init__(self, input_dim, k):
        super().__init__()
        self.register_parameter("weight", nn.Parameter(torch.rand(k, input_dim)))
        self.smax = partial(entmax15, dim=-1)

    def forward(self, x):
        mask = self.smax(self.weight)
        masked_x = torch.einsum("nd,bd->bnd", mask, x)  # [B, k, D]
        return masked_x


class AbstractLayer(nn.Module):
    def __init__(self, base_input_dim, base_output_dim, k, virtual_batch_size, bias=True):
        super().__init__()
        self.masker = LearnableLocality(input_dim=base_input_dim, k=k)
        self.fc = nn.Conv1d(
            base_input_dim * k,
            2 * k * base_output_dim,
            kernel_size=1,
            groups=k,
            bias=bias,
        )
        initialize_glu(self.fc, input_dim=base_input_dim * k, output_dim=2 * k * base_output_dim)
        self.bn = GBN(2 * base_output_dim * k, virtual_batch_size)
        self.k = k
        self.base_output_dim = base_output_dim

    def forward(self, x):
        b = x.size(0)
        x = self.masker(x)  # [B, D] -> [B, k, D]
        x = self.fc(x.view(b, -1, 1))  # [B, k, D] -> [B, k * D, 1] -> [B, k * (2 * D'), 1]
        x = self.bn(x)
        chunks = x.chunk(self.k, 1)  # k * [B, 2 * D', 1]
        x = sum(
            [
                F.relu(torch.sigmoid(x_[:, : self.base_output_dim, :]) * x_[:, self.base_output_dim :, :])
                for x_ in chunks
            ]
        )  # k * [B, D', 1] -> [B, D', 1]
        return x.squeeze(-1)


class BasicBlock(nn.Module):
    def __init__(
        self,
        input_dim,
        abstlay_dim_1,
        abstlay_dim_2,
        k,
        virtual_batch_size,
        fix_input_dim,
        drop_rate,
        block_activation,
    ):
        super().__init__()
        self.conv1 = AbstractLayer(input_dim, abstlay_dim_1, k, virtual_batch_size)
        self.conv2 = AbstractLayer(abstlay_dim_1, abstlay_dim_2, k, virtual_batch_size)

        self.downsample = nn.Sequential(
            nn.Dropout(drop_rate),
            AbstractLayer(fix_input_dim, abstlay_dim_2, k, virtual_batch_size),
        )
        self.block_activation = block_activation

    def forward(self, x, pre_out=None):
        if pre_out is None:
            pre_out = x
        out = self.conv1(pre_out)
        out = self.conv2(out)
        identity = self.downsample(x)
        out += identity
        return self.block_activation(out)



from .conversations import SeparatorStyle   
def build_inputs_labels(self, tokenizer, multi_rounds=True):
    system_prompt = self.system_template.format(system_message=self.system_message)
    encode_fn = lambda text: tokenizer.encode(text,add_special_tokens=False)
    ignore_fn = lambda arr:[-100 for _ in arr]
    eos = [tokenizer.eos_token_id]
    
    if self.sep_style == SeparatorStyle.ADD_COLON_SINGLE:
        inputs = encode_fn(system_prompt) + encode_fn(self.sep)
        labels = ignore_fn(inputs)
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) for x in [role + ": ",message,self.sep]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + ":")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels
        
    elif self.sep_style == SeparatorStyle.ADD_COLON_TWO:
        seps = [self.sep, self.sep2]
        inputs = encode_fn(system_prompt) + eoncde_fn(seps[0])
        labels = ignore_fn(inputs)
        for i, (role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ": ",message,seps[i % 2]]]
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + ":")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels
        
    elif self.sep_style == SeparatorStyle.ADD_COLON_SPACE_SINGLE:
        inputs = encode_fn(system_prompt) + eoncde_fn(self.sep)
        labels = ignore_fn(inputs)
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ": ",message,self.sep]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
                    
            else:
                pre = encode_fn(role + ": ")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels
        
    elif self.sep_style == SeparatorStyle.ADD_NEW_LINE_SINGLE:        
        inputs = [] if system_prompt == "" else encode_fn(
            system_prompt) + eoncde_fn(self.sep)
        labels = ignore_fn(inputs)
        
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + "\n",message,self.sep]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
    
            else:
                pre = encode_fn(role + "\n")
                inputs += pre
                labels += ignore_fn(pre)
                
        return inputs,labels
        
    elif self.sep_style == SeparatorStyle.NO_COLON_SINGLE:        
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role,message,self.sep]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role)
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels
        
    elif self.sep_style == SeparatorStyle.NO_COLON_TWO:
        seps = [self.sep, self.sep2]
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        for i, (role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role,message,seps[i % 2]]]
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role)
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels

    elif self.sep_style == SeparatorStyle.RWKV:
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        for i, (role, message) in enumerate(self.messages):
            if message:
                message = message.replace("\r\n", "\n").replace("\n\n", "\n")
                pre,msg,post = [encode_fn(x) for x in [role+': ',message,"\n\n"]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
                    
            else:
                pre = encode_fn(role + ":")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels

    elif self.sep_style == SeparatorStyle.LLAMA2:
        seps = [self.sep, self.sep2]
        inputs,labels = [],[]
        for i, (role, message) in enumerate(self.messages):
            if message:
                if i == 0:
                    tokens = encode_fn(system_prompt + message + " ")
                    inputs += tokens
                    labels += ignore_fn(tokens)
                else:
                    pre,msg,post = [encode_fn(x) 
                                for x in [role + " ",message,seps[i % 2]]]
                    
                    if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                        inputs += (pre + msg + post)
                        labels+= ignore_fn(pre + msg + post)
                    else:
                        inputs += (pre + msg + eos + post)
                        labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role)
                inputs += pre
                labels += ignore_fn(pre)
        return inputs, labels
        
    elif self.sep_style == SeparatorStyle.CHATGLM:
        round_add_n = 1 if self.name == "chatglm2" else 0
        inputs,labels = [],[]
        if system_prompt:
            inputs += encode_fn(system_prompt + self.sep)
            labels += ignore_fn(inputs)
        for i, (role, message) in enumerate(self.messages):
            if i % 2 == 0:
                tokens = encode_fn(f"[Round {i//2 + round_add_n}]{self.sep}")
                inputs += tokens
                labels += ignore_fn(tokens)
                
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [f"{role}：",message,self.sep]]
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(f"{role}：")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels

    
    elif self.sep_style == SeparatorStyle.CHATML:
        inputs,labels = [],[]
        if system_prompt:
            inputs += encode_fn(system_prompt + self.sep + "\n")
            labels += ignore_fn(inputs)
            
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + "\n", message, self.sep + "\n"]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + "\n")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels

    
    elif self.sep_style == SeparatorStyle.CHATINTERN:
        seps = [self.sep, self.sep2]
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        
        for i, (role, message) in enumerate(self.messages):
            if i % 2 == 0:
                tokens = encode_fn("<s>")
                inputs += tokens
                labels += ignore_fn(tokens)
                
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ":", message, seps[i % 2] + "\n"]]
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg +  post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + ":")
                inputs += pre
                labels += ignore_fn(pre)
                
        return inputs,labels

    
    elif self.sep_style == SeparatorStyle.DOLLY:
        seps = [self.sep, self.sep2]
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        for i, (role, message) in enumerate(self.messages):
            if message:                
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ":\n", message, seps[i % 2]]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
                    
                if i % 2 == 1:
                    tokens = encode_fn("\n\n")
                    inputs += tokens
                    labels += ignore_fn(tokens)
            else:
                pre = encode_fn(role + ":\n")
                inputs += pre
                labels += ignore_fn(pre)
                
        return inputs,labels

    
    elif self.sep_style == SeparatorStyle.PHOENIX:
        inputs = encode_fn(system_prompt)
        labels = ignore_fn(inputs)
        
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ": " + "<s>", message, "</s>"]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + ": " + "<s>")
                inputs += pre
                labels += ignore_fn(pre)
                
        return inputs,labels

    
    elif self.sep_style == SeparatorStyle.ROBIN:
        inputs = encode_fn(system_prompt + self.sep)
        labels = ignore_fn(inputs)
        
        for i,(role, message) in enumerate(self.messages):
            if message:
                pre,msg,post = [encode_fn(x) 
                                for x in [role + ":\n", message, self.sep]]
                
                if role==self.roles[0] or (not multi_rounds and i<len(self.messages)-1):
                    inputs += (pre + msg + post)
                    labels+= ignore_fn(pre + msg + post)
                else:
                    inputs += (pre + msg + eos + post)
                    labels+= (ignore_fn(pre) + msg + eos + ignore_fn(post))
            else:
                pre = encode_fn(role + ":\n")
                inputs += pre
                labels += ignore_fn(pre)
        return inputs,labels
    else:
        raise ValueError(f"Invalid style: {self.sep_style}")


# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
import warnings

import torch
import torch.nn as nn
from omegaconf import DictConfig

from ..base_model import BaseModel
from ..common import activations
from ..common.layers import Lambda
from ..common.layers import Embedding1dLayer
from .architecture_blocks import DenseODSTBlock

class NODEBackbone(nn.Module):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__()
        self.hparams = config
        # self.hparams.output_dim = (0 if self.hparams.output_dim is None else self.hparams.output_dim)
        # For SSL cases where output_dim will be None
        self._build_network()

    def _build_network(self):
        self.hparams.node_input_dim = self.hparams.continuous_dim + self.hparams.embedded_cat_dim
        self.dense_block = DenseODSTBlock(
            input_dim=self.hparams.node_input_dim,
            num_trees=self.hparams.num_trees,
            num_layers=self.hparams.num_layers,
            tree_output_dim=self.hparams.output_dim + self.hparams.additional_tree_output_dim,
            max_features=self.hparams.max_features,
            input_dropout=self.hparams.input_dropout,
            depth=self.hparams.depth,
            choice_function=getattr(activations, self.hparams.choice_function),
            bin_function=getattr(activations, self.hparams.bin_function),
            initialize_response_=getattr(nn.init, self.hparams.initialize_response + "_"),
            initialize_selection_logits_=getattr(nn.init, self.hparams.initialize_selection_logits + "_"),
            threshold_init_beta=self.hparams.threshold_init_beta,
            threshold_init_cutoff=self.hparams.threshold_init_cutoff,
        )
        self.output_dim = self.hparams.output_dim + self.hparams.additional_tree_output_dim

    def _build_embedding_layer(self):
        embedding = Embedding1dLayer(
            continuous_dim=self.hparams.continuous_dim,
            categorical_embedding_dims=self.hparams.embedding_dims,
            embedding_dropout=self.hparams.embedding_dropout,
            batch_norm_continuous_input=self.hparams.batch_norm_continuous_input,
            virtual_batch_size=self.hparams.virtual_batch_size,
        )
        return embedding

    def forward(self, x: torch.Tensor):
        x = self.dense_block(x)
        return x


class NODEModel(BaseModel):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__(config, **kwargs)

    def subset(self, x):
        return x[..., : self.hparams.output_dim].mean(dim=-2)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head

    def _build_network(self):
        self._backbone = NODEBackbone(self.hparams)
        # Embedding Layer
        self._embedding_layer = self._backbone._build_embedding_layer()
        # average first n channels of every tree, where n is the number of output targets for regression
        # and number of classes for classification
        # Not using config head because NODE has a specific head
        warnings.warn("Ignoring head config because NODE has a specific head which subsets the tree outputs")
        self._head = Lambda(self.subset)



import sys,datetime
from tqdm import tqdm
from copy import deepcopy
import numpy as np
import pandas as pd
import torch

"""
@author : lyhue1991, zhangyu
@description : keras core code
"""
class StepRunner:
    def __init__(self, net, loss_fn, accelerator=None, stage="train", metrics_dict=None,
                 optimizer=None, lr_scheduler=None, **kwargs):
        """
        Initialize the StepRunner object

        Args:
            net: Neural network model
            loss_fn: Loss function
            accelerator: Hardware accelerator (e.g., GPU，CPU)
            stage: Training or evaluation stage
            metrics_dict: Dictionary of metrics functions
            optimizer: Optimizer for training
            lr_scheduler: Learning rate scheduler
            **kwargs: Additional keyword arguments
        """
        self.net, self.loss_fn, self.metrics_dict, self.stage = net, loss_fn, metrics_dict, stage
        self.optimizer, self.lr_scheduler = optimizer, lr_scheduler
        self.kwargs = kwargs
        self.accelerator = accelerator

        # Set the network to training mode during the training stage, and evaluation mode otherwise
        if self.stage == 'train':
            self.net.train()
        else:
            self.net.eval()

    def __call__(self, batch):
        """
        Perform a training or evaluation step.

        Args:
            batch: Input data batch.

        Returns:
            Tuple of dictionaries containing step losses and step metrics.
        """
        features, labels = batch

        # Compute loss
        with self.accelerator.autocast():
            preds = self.net(features)
            loss = self.loss_fn(preds, labels)

        # Backward pass and optimization (only during training)
        if self.stage == "train" and self.optimizer is not None:
            self.accelerator.backward(loss)

            # Clip gradients if synchronization is enabled
            if self.accelerator.sync_gradients:
                self.accelerator.clip_grad_norm_(self.net.parameters(), 1.0)

            self.optimizer.step()

            # Adjust learning rate if a scheduler is provided
            if self.lr_scheduler is not None:
                self.lr_scheduler.step()

            # Zero gradients for the next iteration
            self.optimizer.zero_grad()

        # Gather loss, predictions, and labels using the accelerator
        all_loss = self.accelerator.gather(loss).sum()
        all_preds = self.accelerator.gather(preds)
        all_labels = self.accelerator.gather(labels)

        # Compute and gather additional metrics
        step_losses = {self.stage + "_loss": all_loss.item()}
        step_metrics = {self.stage + "_" + name: metric_fn(all_preds, all_labels).item()
                        for name, metric_fn in self.metrics_dict.items()}

        # Include learning rate in metrics if available
        if self.stage=="train":
            if self.optimizer is not None:
                step_metrics['lr'] = self.optimizer.state_dict()['param_groups'][0]['lr']
            else:
                step_metrics['lr'] = 0.0
                
        return step_losses, step_metrics


class EpochRunner:
    def __init__(self, step_runner, quiet=False):
        """
        Initialize the EpochRunner object

        Args:
            step_runner: StepRunner object for handling individual training or evaluation steps
            quiet: Flag to control whether to display progress bar and logs
        """
        self.step_runner = step_runner
        self.stage = step_runner.stage
        self.accelerator = step_runner.accelerator
        self.net = step_runner.net
        self.quiet = quiet

    def __call__(self, dataloader):
        """
        Perform a complete epoch of training or evaluation

        Args:
            dataloader: DataLoader providing batches of data for the epoch

        Returns:
            Dictionary containing aggregated epoch losses and metrics
        """
        # Determine the size of the dataset
        n = dataloader.size if hasattr(dataloader, 'size') else len(dataloader)

        # Initialize tqdm progress bar
        loop = tqdm(enumerate(dataloader, start=1),
                    total=n,
                    file=sys.stdout,
                    disable=not self.accelerator.is_local_main_process or self.quiet,
                    ncols=100
                    )
        epoch_losses = {}

        for step, batch in loop:
            # Perform a step with the provided StepRunner
            with self.accelerator.accumulate(self.net):
                step_losses, step_metrics = self.step_runner(batch)
                step_log = dict(step_losses, **step_metrics)

                # Accumulate step losses for computing epoch losses
                for k, v in step_losses.items():
                    epoch_losses[k] = epoch_losses.get(k, 0.0) + v

                # Update progress bar during the epoch
                if step < n:
                    loop.set_postfix(**step_log)

                    if hasattr(self, 'progress') and self.accelerator.is_local_main_process:
                        post_log = dict(**{'i': step, 'n': n}, **step_log)
                        self.progress.set_postfix(**post_log)

                # Compute and display epoch-level metrics at the end of the epoch
                elif step == n:
                    epoch_metrics = step_metrics
                    epoch_metrics.update({self.stage + "_" + name: metric_fn.compute().item()
                                          for name, metric_fn in self.step_runner.metrics_dict.items()})
                    epoch_losses = {k: v / step for k, v in epoch_losses.items()}
                    epoch_log = dict(epoch_losses, **epoch_metrics)
                    loop.set_postfix(**epoch_log)

                    # Update progress bar if available
                    if hasattr(self, 'progress') and self.accelerator.is_local_main_process:
                        post_log = dict(**{'i': step, 'n': n}, **epoch_log)
                        self.progress.set_postfix(**post_log)

                    # Reset stateful metrics for the next epoch
                    for name, metric_fn in self.step_runner.metrics_dict.items():
                        metric_fn.reset()
                else:
                    break

        return epoch_log

class KerasModel(torch.nn.Module):
    StepRunner, EpochRunner = StepRunner, EpochRunner

    def __init__(self, net, loss_fn, metrics_dict=None, optimizer=None, lr_scheduler=None, **kwargs):
        """
        Initialize the KerasModel.

        Args:
            net: Neural network model
            loss_fn: Loss function
            metrics_dict: Dictionary of metrics functions
            optimizer: Optimizer for training
            lr_scheduler: Learning rate scheduler
            **kwargs: Additional keyword arguments
        """
        super().__init__()
        self.net, self.loss_fn, self.metrics_dict = net, loss_fn, torch.nn.ModuleDict(metrics_dict)
        self.optimizer = optimizer if optimizer is not None else torch.optim.Adam(
            self.net.parameters(), lr=3e-4)
        self.lr_scheduler = lr_scheduler
        self.kwargs = kwargs
        self.from_scratch = True

    def save_ckpt(self, ckpt_path=None, accelerator=None):
        """
        Save the model checkpoint

        Args:
            ckpt_path: Path to save the checkpoint
            accelerator: Hardware accelerator (e.g., GPU)
        """
        accelerator = accelerator if accelerator is not None else self.accelerator
        net_dict = accelerator.get_state_dict(self.net)
        accelerator.save(net_dict, ckpt_path if ckpt_path is not None else self.ckpt_path)

    def load_ckpt(self, ckpt_path=None):
        """
        Load the model checkpoint

        Args:
            ckpt_path: Path to the checkpoint
        """
        net_dict = torch.load(ckpt_path if ckpt_path is not None else self.ckpt_path,
                               map_location='cpu', weights_only=True)
        self.net.load_state_dict(net_dict)
        self.from_scratch = False

    def forward(self, x):
        """
        Forward pass through the model

        Args:
            x: Input data

        Returns:
            Model predictions
        """
        return self.net.forward(x)

    def fit(self, train_data, val_data=None, epochs=10, ckpt_path='checkpoint',
            patience=5, monitor="val_loss", mode="min", 
            callbacks=None, plot=True, wandb=False, 
            mixed_precision='no', cpu=False, gradient_accumulation_steps=1):
        """
        Train the model.

        Args:
            train_data: Training data
            val_data: Validation data
            epochs: Number of training epochs
            ckpt_path: Path to save model checkpoints
            patience: Number of epochs with no improvement after which training will be stopped
            monitor: Metric to monitor for early stopping
            mode: 'min' for minimizing the monitor metric, 'max' for maximizing
            callbacks: List of callback functions
            plot: Whether to plot training progress
            wandb: Whether to use WandB for logging
            mixed_precision: Mixed precision training ('no', 'O1', 'O2', 'O3')
            cpu: Use CPU for training
            gradient_accumulation_steps: Number of steps to accumulate gradients before optimizer step

        Returns:
            DataFrame containing training history.
        """
        self.__dict__.update(locals())
        from accelerate import Accelerator
        from torchkeras.utils import colorful, is_jupyter

        self.accelerator = Accelerator(mixed_precision=mixed_precision, cpu=cpu,
                                       gradient_accumulation_steps=gradient_accumulation_steps)

        device = str(self.accelerator.device)
        device_type = '🐌' if 'cpu' in device else ('⚡️' if 'cuda' in device else '🚀')
        self.accelerator.print(
            colorful("<<<<<< " + device_type + " " + device + " is used >>>>>>"))

        self.net, self.loss_fn, self.metrics_dict, self.optimizer, self.lr_scheduler = self.accelerator.prepare(
            self.net, self.loss_fn, self.metrics_dict, self.optimizer, self.lr_scheduler)

        for key in self.kwargs:
            self.kwargs[key] = self.accelerator.prepare(self.kwargs[key])

        train_dataloader, val_dataloader = self.accelerator.prepare(train_data, val_data)
        train_dataloader.size = train_data.size if hasattr(train_data, 'size') else len(train_data)
        train_dataloader.size = min(train_dataloader.size, len(train_dataloader))

        if val_data:
            val_dataloader.size = val_data.size if hasattr(val_data, 'size') else len(val_data)
            val_dataloader.size = min(val_dataloader.size, len(val_dataloader))

        self.history = {}
        callbacks = callbacks if callbacks is not None else []

        
        if bool(plot):
            from torchkeras.kerascallbacks import VisProgress, VisMetric
            callbacks = [VisMetric(), VisProgress()] + callbacks

        if wandb != False:
            from torchkeras.kerascallbacks import WandbCallback
            project = wandb if isinstance(wandb, str) else 'torchkeras'
            callbacks.append(WandbCallback(project=project))

        self.callbacks = [self.accelerator.prepare(x) for x in callbacks]

        if self.accelerator.is_local_main_process:
            [cb.on_fit_start(model=self) for cb in self.callbacks if hasattr(cb, 'on_fit_start')]

        start_epoch = 1 if self.from_scratch else 0
        quiet = bool(plot)
        
        for epoch in range(start_epoch, epochs + 1):
            if not quiet:
                nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                self.accelerator.print("\n" + "==========" * 8 + "%s" % nowtime)
                self.accelerator.print("Epoch {0} / {1}".format(epoch, epochs) + "\n")

            # 1，train -------------------------------------------------
            train_step_runner = self.StepRunner(
                net=self.net,
                loss_fn=self.loss_fn,
                accelerator=self.accelerator,
                stage="train",
                metrics_dict=deepcopy(self.metrics_dict),
                optimizer=self.optimizer if epoch > 0 else None,
                lr_scheduler=self.lr_scheduler if epoch > 0 else None,
                **self.kwargs
            )

            train_epoch_runner = self.EpochRunner(train_step_runner, quiet)
            train_metrics = {'epoch': epoch}
            train_metrics.update(train_epoch_runner(train_dataloader))

            for name, metric in train_metrics.items():
                self.history[name] = self.history.get(name, []) + [metric]

            if self.accelerator.is_local_main_process:
                [cb.on_train_epoch_end(model=self) for cb in self.callbacks
                 if hasattr(cb, 'on_train_epoch_end')]
            # 2，validate -------------------------------------------------
            if val_dataloader is not None:
                val_step_runner = self.StepRunner(
                    net = self.net,
                    loss_fn = self.loss_fn,
                    accelerator = self.accelerator,
                    stage="val",
                    metrics_dict= deepcopy(self.metrics_dict),
                    **self.kwargs
                )
                val_epoch_runner = self.EpochRunner(val_step_runner,quiet)
                with torch.no_grad():
                    val_metrics = val_epoch_runner(val_dataloader)

                for name, metric in val_metrics.items():
                    self.history[name] = self.history.get(name, []) + [metric]
                
            if self.accelerator.is_local_main_process:
                [cb.on_validation_epoch_end(model = self) for cb in self.callbacks 
                 if hasattr(cb,'on_validation_epoch_end')]

            # 3，early-stopping -------------------------------------------------
            self.accelerator.wait_for_everyone()
            arr_scores = self.history[monitor]
            best_score_idx = np.argmax(arr_scores) if mode=="max" else np.argmin(arr_scores)

            if best_score_idx==len(arr_scores)-1 and self.accelerator.is_local_main_process:
                self.save_ckpt(ckpt_path,accelerator = self.accelerator)
                if not quiet:
                    self.accelerator.print(colorful("<<<<<< reach best {0} : {1} >>>>>>".format(
                        monitor,arr_scores[best_score_idx])))

            if len(arr_scores)-best_score_idx>patience:
                break
                
        if self.accelerator.is_local_main_process:   
            dfhistory = pd.DataFrame(self.history)
            [cb.on_fit_end(model = self) for cb in self.callbacks 
                 if hasattr(cb,'on_fit_end')]
            if epoch<epochs:
                self.accelerator.print(colorful(
                        "<<<<<< {} without improvement in {} epoch,""early stopping >>>>>> \n"
                    ).format(monitor,patience))
            self.net = self.accelerator.unwrap_model(self.net)
            self.net.cpu()
            self.load_ckpt(ckpt_path)
            return dfhistory
        
    def evaluate(self, val_data, quiet=False):
        """
        Evaluate the model on validation data

        Args:
            val_data: Validation data
            quiet: Whether to suppress evaluation progress logs
            
        Returns:
            Dictionary of evaluation metrics
        """
        # Ensure accelerator is available or create a new one
        from accelerate import Accelerator
        accelerator = Accelerator() if not hasattr(self, 'accelerator') else self.accelerator

        # Prepare model, loss function, and metrics for evaluation
        self.net, self.loss_fn, self.metrics_dict = accelerator.prepare(
            self.net, self.loss_fn, self.metrics_dict)

        val_data = accelerator.prepare(val_data)

        # Initialize StepRunner for validation
        val_step_runner = self.StepRunner(net=self.net, stage="val",
                                          loss_fn=self.loss_fn, metrics_dict=deepcopy(self.metrics_dict),
                                          accelerator=accelerator)

        # Initialize EpochRunner for validation
        val_epoch_runner = self.EpochRunner(val_step_runner, quiet=quiet)

        # Evaluate on validation data without gradient computation
        with torch.no_grad():
            val_metrics = val_epoch_runner(val_data)

        return val_metrics
    
    def fit_ddp(self, num_processes, train_data,
                val_data=None, epochs=10, ckpt_path='checkpoint',
                patience=5, monitor="val_loss", mode="min", callbacks=None,
                plot=True, wandb=False, mixed_precision='no', 
                cpu=False, gradient_accumulation_steps=1):
        """
        Distributed Data Parallel (DDP) training for the model.

        Args:
            num_processes: Number of processes for DDP
            train_data: Training data
            val_data: Validation data
            epochs: Number of training epochs
            ckpt_path: Path to save model checkpoints
            patience: Number of epochs with no improvement after which training will be stopped
            monitor: Metric to monitor for early stopping
            mode: 'min' for minimizing the monitor metric, 'max' for maximizing
            callbacks: List of callback functions
            plot: Whether to plot training progress
            wandb: Whether to use WandB for logging
            mixed_precision: Mixed precision training ('no', 'O1', 'O2', 'O3')
            cpu: Use CPU for training
            gradient_accumulation_steps: Number of steps to accumulate gradients before optimizer step
        """
        # Import notebook_launcher from accelerate
        from accelerate import notebook_launcher

        # Create a tuple of arguments for the fit method
        args = (train_data, val_data, epochs, ckpt_path, patience, monitor, mode,
                callbacks, plot, wandb, mixed_precision, cpu, gradient_accumulation_steps)

        # Launch the fit method using notebook_launcher
        notebook_launcher(self.fit, args, num_processes=num_processes)
    
    def evaluate_ddp(self, num_processes, val_data, quiet=False):
        """
        Distributed Data Parallel (DDP) evaluation for the model

        Args:
            num_processes: Number of processes for DDP
            val_data: Validation data.
            quiet: Whether to suppress evaluation progress logs

        Returns:
            Dictionary of evaluation metrics
        """
        # Import notebook_launcher from accelerate
        from accelerate import notebook_launcher

        # Create a tuple of arguments for the evaluate method
        args = (val_data, quiet)

        # Launch the evaluate method using notebook_launcher
        notebook_launcher(self.evaluate, args, num_processes=num_processes)


# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
"""Tabnet Model Config."""
from dataclasses import dataclass, field
from typing import List, Optional

from torchkeras.tabular.config import ModelConfig

@dataclass
class TabNetModelConfig(ModelConfig):
    """TabNet: Attentive Interpretable Tabular Learning configuration

    Args:
        n_d (int): Dimension of the prediction  layer (usually between 4 and 64)

        n_a (int): Dimension of the attention  layer (usually between 4 and 64)

        n_steps (int): Number of successive steps in the network (usually between 3 and 10)

        gamma (float): Float above 1, scaling factor for attention updates (usually between 1.0 to 2.0)

        n_independent (int): Number of independent GLU layer in each GLU block (default 2)

        n_shared (int): Number of independent GLU layer in each GLU block (default 2)

        virtual_batch_size (int): Batch size for Ghost Batch Normalization

        mask_type (str): Either 'sparsemax' or 'entmax' : this is the masking function to use. Choices are:
                [`sparsemax`,`entmax`].

        task (str): Specify whether the problem is regression or classification. `backbone` is a task which
                considers the model as a backbone to generate features. Mostly used internally for SSL and related
                tasks. Choices are: [`regression`,`classification`,`backbone`].

        head (Optional[str]): The head to be used for the model. Should be one of the heads defined in
                `pytorch_tabular.models.common.heads`. Defaults to  LinearHead. Choices are:
                [`None`,`LinearHead`,`MixtureDensityHead`].

        head_config (Optional[Dict]): The config as a dict which defines the head. If left empty, will be
                initialized as default linear head.

        embedding_dims (Optional[List]): The dimensions of the embedding for each categorical column as a
                list of tuples (cardinality, embedding_dim). If left empty, will infer using the cardinality of
                the categorical column using the rule min(50, (x + 1) // 2)

        embedding_dropout (float): Dropout to be applied to the Categorical Embedding. Defaults to 0.0

        batch_norm_continuous_input (bool): If True, we will normalize the continuous layer by passing it
                through a BatchNorm layer.

        loss (Optional[str]): The loss function to be applied. By Default, it is MSELoss for regression and
                CrossEntropyLoss for classification. Unless you are sure what you are doing, leave it at MSELoss
                or L1Loss for regression and CrossEntropyLoss for classification
    """

    n_d: int = field(
        default=8,
        metadata={"help": "Dimension of the prediction  layer (usually between 4 and 64)"},
    )
    n_a: int = field(
        default=8,
        metadata={"help": "Dimension of the attention  layer (usually between 4 and 64)"},
    )
    n_steps: int = field(
        default=3,
        metadata={"help": ("Number of successive steps in the network (usually between 3 and 10)")},
    )
    gamma: float = field(
        default=1.3,
        metadata={"help": ("Float above 1, scaling factor for attention updates (usually between" " 1.0 to 2.0)")},
    )
    n_independent: int = field(
        default=2,
        metadata={"help": "Number of independent GLU layer in each GLU block (default 2)"},
    )
    n_shared: int = field(
        default=2,
        metadata={"help": "Number of independent GLU layer in each GLU block (default 2)"},
    )
    virtual_batch_size: int = field(
        default=128,
        metadata={"help": "Batch size for Ghost Batch Normalization"},
    )
    mask_type: str = field(
        default="sparsemax",
        metadata={
            "help": ("Either 'sparsemax' or 'entmax' : this is the masking function to use"),
            "choices": ["sparsemax", "entmax"],
        },
    )
    grouped_features: Optional[List[List[str]]] = field(
        default=None,
        metadata={
            "help": (
                "List of list of feature names to be grouped together. This allows the"
                " model to share it's attention accross feature inside a same group."
                " This can be especially useful when your preprocessing generates"
                " correlated or dependant features: like if you use a TF-IDF or a PCA"
                " on a text column. Note that feature importance will be exactly the"
                " same between features on a same group. Please also note that"
                " embeddings generated for a categorical variable are always inside a"
                " same group."
            )
        },
    )
    _module_src: str = field(default="models.tabnet")
    _model_name: str = field(default="TabNetModel")
    _config_name: str = field(default="TabNetModelConfig")





import numpy as np
import random
import warnings
import inspect
import torch
from typing import Callable, List, Optional, Union
from torch import nn
import torch.distributed as dist
import copy

from transformers import set_seed 
from transformers.generation.utils import GenerateOutput, SampleOutput, logger
from transformers import (
    GenerationConfig,
    GenerationMixin,
    LogitsProcessorList,
    StoppingCriteriaList,
    DisjunctiveConstraint,
    BeamSearchScorer,
    PhrasalConstraint,
    ConstrainedBeamSearchScorer,
    PreTrainedModel,
)
'''
reference @ https://github.com/LowinLi/transformers-stream-generator
'''

class StreamGenerationConfig(GenerationConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.do_stream = kwargs.pop("do_stream", False)

class NewGenerationMixin(GenerationMixin):
    @torch.no_grad()
    def generate(
        self,
        inputs: Optional[torch.Tensor] = None,
        generation_config: Optional[StreamGenerationConfig] = None,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        prefix_allowed_tokens_fn: Optional[
            Callable[[int, torch.Tensor], List[int]]
        ] = None,
        synced_gpus: Optional[bool] = False,
        seed=0,
        **kwargs,
    ) -> Union[GenerateOutput, torch.LongTensor]:
        r"""

        Generates sequences of token ids for models with a language modeling head.

        <Tip warning={true}>

        Most generation-controlling parameters are set in `generation_config` which, if not passed, will be set to the
        model's default generation configuration. You can override any `generation_config` by passing the corresponding
        parameters to generate(), e.g. `.generate(inputs, num_beams=4, do_sample=True)`.

        For an overview of generation strategies and code examples, check out the [following
        guide](./generation_strategies).

        </Tip>

        Parameters:
            inputs (`torch.Tensor` of varying shape depending on the modality, *optional*):
                The sequence used as a prompt for the generation or as model inputs to the encoder. If `None` the
                method initializes it with `bos_token_id` and a batch size of 1. For decoder-only models `inputs`
                should of in the format of `input_ids`. For encoder-decoder models *inputs* can represent any of
                `input_ids`, `input_values`, `input_features`, or `pixel_values`.
            generation_config (`~generation.GenerationConfig`, *optional*):
                The generation configuration to be used as base parametrization for the generation call. `**kwargs`
                passed to generate matching the attributes of `generation_config` will override them. If
                `generation_config` is not provided, the default will be used, which had the following loading
                priority: 1) from the `generation_config.json` model file, if it exists; 2) from the model
                configuration. Please note that unspecified parameters will inherit [`~generation.GenerationConfig`]'s
                default values, whose documentation should be checked to parameterize generation.
            logits_processor (`LogitsProcessorList`, *optional*):
                Custom logits processors that complement the default logits processors built from arguments and
                generation config. If a logit processor is passed that is already created with the arguments or a
                generation config an error is thrown. This feature is intended for advanced users.
            stopping_criteria (`StoppingCriteriaList`, *optional*):
                Custom stopping criteria that complement the default stopping criteria built from arguments and a
                generation config. If a stopping criteria is passed that is already created with the arguments or a
                generation config an error is thrown. This feature is intended for advanced users.
            prefix_allowed_tokens_fn (`Callable[[int, torch.Tensor], List[int]]`, *optional*):
                If provided, this function constraints the beam search to allowed tokens only at each step. If not
                provided no constraint is applied. This function takes 2 arguments: the batch ID `batch_id` and
                `input_ids`. It has to return a list with the allowed tokens for the next generation step conditioned
                on the batch ID `batch_id` and the previously generated tokens `inputs_ids`. This argument is useful
                for constrained generation conditioned on the prefix, as described in [Autoregressive Entity
                Retrieval](https://arxiv.org/abs/2010.00904).
            synced_gpus (`bool`, *optional*, defaults to `False`):
                Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
            kwargs:
                Ad hoc parametrization of `generate_config` and/or additional model-specific kwargs that will be
                forwarded to the `forward` function of the model. If the model is an encoder-decoder model, encoder
                specific kwargs should not be prefixed and decoder specific kwargs should be prefixed with *decoder_*.

        Return:
            [`~utils.ModelOutput`] or `torch.LongTensor`: A [`~utils.ModelOutput`] (if `return_dict_in_generate=True`
            or when `config.return_dict_in_generate=True`) or a `torch.FloatTensor`.

                If the model is *not* an encoder-decoder model (`model.config.is_encoder_decoder=False`), the possible
                [`~utils.ModelOutput`] types are:

                    - [`~generation.GreedySearchDecoderOnlyOutput`],
                    - [`~generation.SampleDecoderOnlyOutput`],
                    - [`~generation.BeamSearchDecoderOnlyOutput`],
                    - [`~generation.BeamSampleDecoderOnlyOutput`]

                If the model is an encoder-decoder model (`model.config.is_encoder_decoder=True`), the possible
                [`~utils.ModelOutput`] types are:

                    - [`~generation.GreedySearchEncoderDecoderOutput`],
                    - [`~generation.SampleEncoderDecoderOutput`],
                    - [`~generation.BeamSearchEncoderDecoderOutput`],
                    - [`~generation.BeamSampleEncoderDecoderOutput`]
        """
        set_seed(seed)
        # 1. Handle `generation_config` and kwargs that might update it, and validate the `.generate()` call
        self._validate_model_class()

        # priority: `generation_config` argument > `model.generation_config` (the default generation config)
        if generation_config is None:
            # legacy: users may modify the model configuration to control generation -- update the generation config
            # model attribute accordingly, if it was created from the model config
            if self.generation_config._from_model_config:
                new_generation_config = StreamGenerationConfig.from_model_config(
                    self.config
                )
                if new_generation_config != self.generation_config:
                    warnings.warn(
                        "You have modified the pretrained model configuration to control generation. This is a"
                        " deprecated strategy to control generation and will be removed soon, in a future version."
                        " Please use a generation configuration file (see"
                        " https://huggingface.co/docs/transformers/main_classes/text_generation)"
                    )
                    self.generation_config = new_generation_config
            generation_config = self.generation_config

        generation_config = copy.deepcopy(generation_config)
        model_kwargs = generation_config.update(
            **kwargs
        )  # All unused kwargs must be model kwargs
        # self._validate_model_kwargs(model_kwargs.copy())

        # 2. Set generation parameters if not already defined
        logits_processor = (
            logits_processor if logits_processor is not None else LogitsProcessorList()
        )
        stopping_criteria = (
            stopping_criteria
            if stopping_criteria is not None
            else StoppingCriteriaList()
        )

        if (
            generation_config.pad_token_id is None
            and generation_config.eos_token_id is not None
        ):
            if model_kwargs.get("attention_mask", None) is None:
                logger.warning(
                    "The attention mask and the pad token id were not set. As a consequence, you may observe "
                    "unexpected behavior. Please pass your input's `attention_mask` to obtain reliable results."
                )
            eos_token_id = generation_config.eos_token_id
            if isinstance(eos_token_id, list):
                eos_token_id = eos_token_id[0]
            logger.warning(
                f"Setting `pad_token_id` to `eos_token_id`:{eos_token_id} for open-end generation."
            )
            generation_config.pad_token_id = eos_token_id

        # 3. Define model inputs
        # inputs_tensor has to be defined
        # model_input_name is defined if model-specific keyword input is passed
        # otherwise model_input_name is None
        # all model-specific keyword inputs are removed from `model_kwargs`
        inputs_tensor, model_input_name, model_kwargs = self._prepare_model_inputs(
            inputs, generation_config.bos_token_id, model_kwargs
        )
        batch_size = inputs_tensor.shape[0]

        # 4. Define other model kwargs
        model_kwargs["output_attentions"] = generation_config.output_attentions
        model_kwargs["output_hidden_states"] = generation_config.output_hidden_states
        model_kwargs["use_cache"] = generation_config.use_cache

        accepts_attention_mask = "attention_mask" in set(
            inspect.signature(self.forward).parameters.keys()
        )
        requires_attention_mask = "encoder_outputs" not in model_kwargs

        if (
            model_kwargs.get("attention_mask", None) is None
            and requires_attention_mask
            and accepts_attention_mask
        ):
            model_kwargs[
                "attention_mask"
            ] = self._prepare_attention_mask_for_generation(
                inputs_tensor,
                generation_config.pad_token_id,
                generation_config.eos_token_id,
            )

        # decoder-only models should use left-padding for generation
        if not self.config.is_encoder_decoder:
            if (
                generation_config.pad_token_id is not None
                and torch.sum(inputs_tensor[:, -1] == generation_config.pad_token_id)
                > 0
            ):
                logger.warning(
                    "A decoder-only architecture is being used, but right-padding was detected! For correct "
                    "generation results, please set `padding_side='left'` when initializing the tokenizer."
                )

        if self.config.is_encoder_decoder and "encoder_outputs" not in model_kwargs:
            # if model is encoder decoder encoder_outputs are created
            # and added to `model_kwargs`
            model_kwargs = self._prepare_encoder_decoder_kwargs_for_generation(
                inputs_tensor, model_kwargs, model_input_name
            )

        # 5. Prepare `input_ids` which will be used for auto-regressive generation
        if self.config.is_encoder_decoder:
            input_ids = self._prepare_decoder_input_ids_for_generation(
                batch_size,
                decoder_start_token_id=generation_config.decoder_start_token_id,
                bos_token_id=generation_config.bos_token_id,
                model_kwargs=model_kwargs,
                device=inputs_tensor.device,
            )
        else:
            # if decoder-only then inputs_tensor has to be `input_ids`
            input_ids = inputs_tensor

        # 6. Prepare `max_length` depending on other stopping criteria.
        input_ids_seq_length = input_ids.shape[-1]
        has_default_max_length = (
            kwargs.get("max_length") is None
            and generation_config.max_length is not None
        )
        if has_default_max_length and generation_config.max_new_tokens is None:
            warnings.warn(
                "Neither `max_length` nor `max_new_tokens` has been set, `max_length` will default to"
                f" {generation_config.max_length} (`generation_config.max_length`). Controlling `max_length` via the"
                " config is deprecated and `max_length` will be removed from the config in v5 of Transformers -- we"
                " recommend using `max_new_tokens` to control the maximum length of the generation.",
                UserWarning,
            )
        elif has_default_max_length and generation_config.max_new_tokens is not None:
            generation_config.max_length = (
                generation_config.max_new_tokens + input_ids_seq_length
            )
        elif (
            not has_default_max_length and generation_config.max_new_tokens is not None
        ):
            raise ValueError(
                "Both `max_new_tokens` and `max_length` have been set but they serve the same purpose -- setting a"
                " limit to the generated output length. Remove one of those arguments. Please refer to the"
                " documentation for more information. "
                "(https://huggingface.co/docs/transformers/main/en/main_classes/text_generation)"
            )

        if (
            generation_config.min_length is not None
            and generation_config.min_length > generation_config.max_length
        ):
            raise ValueError(
                f"Unfeasible length constraints: the minimum length ({generation_config.min_length}) is larger than"
                f" the maximum length ({generation_config.max_length})"
            )
        if input_ids_seq_length >= generation_config.max_length:
            input_ids_string = (
                "decoder_input_ids" if self.config.is_encoder_decoder else "input_ids"
            )
            logger.warning(
                f"Input length of {input_ids_string} is {input_ids_seq_length}, but `max_length` is set to"
                f" {generation_config.max_length}. This can lead to unexpected behavior. You should consider"
                " increasing `max_new_tokens`."
            )

        # 7. determine generation mode
        is_constraint_gen_mode = (
            generation_config.constraints is not None
            or generation_config.force_words_ids is not None
        )

        is_contrastive_search_gen_mode = (
            generation_config.top_k is not None
            and generation_config.top_k > 1
            and generation_config.do_sample is False
            and generation_config.penalty_alpha is not None
            and generation_config.penalty_alpha > 0
        )

        is_greedy_gen_mode = (
            (generation_config.num_beams == 1)
            and (generation_config.num_beam_groups == 1)
            and generation_config.do_sample is False
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )
        is_sample_gen_mode = (
            (generation_config.num_beams == 1)
            and (generation_config.num_beam_groups == 1)
            and generation_config.do_sample is True
            and generation_config.do_stream is False
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )
        is_sample_gen_stream_mode = (
            (generation_config.num_beams == 1)
            and (generation_config.num_beam_groups == 1)
            and generation_config.do_stream is True
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )
        is_beam_gen_mode = (
            (generation_config.num_beams > 1)
            and (generation_config.num_beam_groups == 1)
            and generation_config.do_sample is False
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )
        is_beam_sample_gen_mode = (
            (generation_config.num_beams > 1)
            and (generation_config.num_beam_groups == 1)
            and generation_config.do_sample is True
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )
        is_group_beam_gen_mode = (
            (generation_config.num_beams > 1)
            and (generation_config.num_beam_groups > 1)
            and not is_constraint_gen_mode
            and not is_contrastive_search_gen_mode
        )

        if generation_config.num_beam_groups > generation_config.num_beams:
            raise ValueError(
                "`num_beam_groups` has to be smaller or equal to `num_beams`"
            )
        if is_group_beam_gen_mode and generation_config.do_sample is True:
            raise ValueError(
                "Diverse beam search cannot be used in sampling mode. Make sure that `do_sample` is set to `False`."
            )

        if self.device.type != input_ids.device.type:
            warnings.warn(
                "You are calling .generate() with the `input_ids` being on a device type different"
                f" than your model's device. `input_ids` is on {input_ids.device.type}, whereas the model"
                f" is on {self.device.type}. You may experience unexpected behaviors or slower generation."
                " Please make sure that you have put `input_ids` to the"
                f" correct device by calling for example input_ids = input_ids.to('{self.device.type}') before"
                " running `.generate()`.",
                UserWarning,
            )
        # 8. prepare distribution pre_processing samplers
        logits_processor = self._get_logits_processor(
            generation_config=generation_config,
            input_ids_seq_length=input_ids_seq_length,
            encoder_input_ids=inputs_tensor,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            logits_processor=logits_processor,
        )

        # 9. prepare stopping criteria
        stopping_criteria = self._get_stopping_criteria(
            generation_config=generation_config, stopping_criteria=stopping_criteria
        )
        # 10. go into different generation modes
        if is_greedy_gen_mode:
            if generation_config.num_return_sequences > 1:
                raise ValueError(
                    f"num_return_sequences has to be 1, but is {generation_config.num_return_sequences} when doing"
                    " greedy search."
                )

            # 11. run greedy search
            return self.greedy_search(
                input_ids,
                logits_processor=logits_processor,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

        elif is_contrastive_search_gen_mode:
            if generation_config.num_return_sequences > 1:
                raise ValueError(
                    f"num_return_sequences has to be 1, but is {generation_config.num_return_sequences} when doing"
                    " contrastive search."
                )

            return self.contrastive_search(
                input_ids,
                top_k=generation_config.top_k,
                penalty_alpha=generation_config.penalty_alpha,
                logits_processor=logits_processor,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

        elif is_sample_gen_mode:
            # 11. prepare logits warper
            logits_warper = self._get_logits_warper(generation_config)

            # 12. expand input_ids with `num_return_sequences` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_return_sequences,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )

            # 13. run sample
            return self.sample(
                input_ids,
                logits_processor=logits_processor,
                logits_warper=logits_warper,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )
        elif is_sample_gen_stream_mode:
            # 11. prepare logits warper
            logits_warper = self._get_logits_warper(generation_config)

            # 12. expand input_ids with `num_return_sequences` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_return_sequences,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )

            # 13. run sample
            return self.sample_stream(
                input_ids,
                logits_processor=logits_processor,
                logits_warper=logits_warper,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )
        elif is_beam_gen_mode:
            if generation_config.num_return_sequences > generation_config.num_beams:
                raise ValueError(
                    "`num_return_sequences` has to be smaller or equal to `num_beams`."
                )

            if stopping_criteria.max_length is None:
                raise ValueError(
                    "`max_length` needs to be a stopping_criteria for now."
                )

            # 11. prepare beam search scorer
            beam_scorer = BeamSearchScorer(
                batch_size=batch_size,
                num_beams=generation_config.num_beams,
                device=inputs_tensor.device,
                length_penalty=generation_config.length_penalty,
                do_early_stopping=generation_config.early_stopping,
                num_beam_hyps_to_keep=generation_config.num_return_sequences,
            )
            # 12. interleave input_ids with `num_beams` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_beams,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )
            # 13. run beam search
            return self.beam_search(
                input_ids,
                beam_scorer,
                logits_processor=logits_processor,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

        elif is_beam_sample_gen_mode:
            # 11. prepare logits warper
            logits_warper = self._get_logits_warper(generation_config)

            if stopping_criteria.max_length is None:
                raise ValueError(
                    "`max_length` needs to be a stopping_criteria for now."
                )
            # 12. prepare beam search scorer
            beam_scorer = BeamSearchScorer(
                batch_size=batch_size * generation_config.num_return_sequences,
                num_beams=generation_config.num_beams,
                device=inputs_tensor.device,
                length_penalty=generation_config.length_penalty,
                do_early_stopping=generation_config.early_stopping,
            )

            # 13. interleave input_ids with `num_beams` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_beams
                * generation_config.num_return_sequences,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )

            # 14. run beam sample
            return self.beam_sample(
                input_ids,
                beam_scorer,
                logits_processor=logits_processor,
                logits_warper=logits_warper,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

        elif is_group_beam_gen_mode:
            if generation_config.num_return_sequences > generation_config.num_beams:
                raise ValueError(
                    "`num_return_sequences` has to be smaller or equal to `num_beams`."
                )

            if generation_config.num_beams % generation_config.num_beam_groups != 0:
                raise ValueError(
                    "`num_beams` should be divisible by `num_beam_groups` for group beam search."
                )

            if stopping_criteria.max_length is None:
                raise ValueError(
                    "`max_length` needs to be a stopping_criteria for now."
                )

            has_default_typical_p = (
                kwargs.get("typical_p") is None and generation_config.typical_p == 1.0
            )
            if not has_default_typical_p:
                raise ValueError(
                    "Decoder argument `typical_p` is not supported with beam groups."
                )

            # 11. prepare beam search scorer
            beam_scorer = BeamSearchScorer(
                batch_size=batch_size,
                num_beams=generation_config.num_beams,
                max_length=stopping_criteria.max_length,
                device=inputs_tensor.device,
                length_penalty=generation_config.length_penalty,
                do_early_stopping=generation_config.early_stopping,
                num_beam_hyps_to_keep=generation_config.num_return_sequences,
                num_beam_groups=generation_config.num_beam_groups,
            )
            # 12. interleave input_ids with `num_beams` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_beams,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )
            # 13. run beam search
            return self.group_beam_search(
                input_ids,
                beam_scorer,
                logits_processor=logits_processor,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

        elif is_constraint_gen_mode:
            if generation_config.num_return_sequences > generation_config.num_beams:
                raise ValueError(
                    "`num_return_sequences` has to be smaller or equal to `num_beams`."
                )

            if stopping_criteria.max_length is None:
                raise ValueError(
                    "`max_length` needs to be a stopping_criteria for now."
                )

            if generation_config.num_beams <= 1:
                raise ValueError(
                    "`num_beams` needs to be greater than 1 for constrained generation."
                )

            if generation_config.do_sample:
                raise ValueError(
                    "`do_sample` needs to be false for constrained generation."
                )

            if (
                generation_config.num_beam_groups is not None
                and generation_config.num_beam_groups > 1
            ):
                raise ValueError(
                    "`num_beam_groups` not supported yet for constrained generation."
                )

            final_constraints = []
            if generation_config.constraints is not None:
                final_constraints = generation_config.constraints

            if generation_config.force_words_ids is not None:

                def typeerror():
                    raise ValueError(
                        "`force_words_ids` has to either be a `List[List[List[int]]]` or `List[List[int]]`"
                        f"of positive integers, but is {generation_config.force_words_ids}."
                    )

                if (
                    not isinstance(generation_config.force_words_ids, list)
                    or len(generation_config.force_words_ids) == 0
                ):
                    typeerror()

                for word_ids in generation_config.force_words_ids:
                    if isinstance(word_ids[0], list):
                        if not isinstance(word_ids, list) or len(word_ids) == 0:
                            typeerror()
                        if any(
                            not isinstance(token_ids, list) for token_ids in word_ids
                        ):
                            typeerror()
                        if any(
                            any(
                                (not isinstance(token_id, int) or token_id < 0)
                                for token_id in token_ids
                            )
                            for token_ids in word_ids
                        ):
                            typeerror()

                        constraint = DisjunctiveConstraint(word_ids)
                    else:
                        if not isinstance(word_ids, list) or len(word_ids) == 0:
                            typeerror()
                        if any(
                            (not isinstance(token_id, int) or token_id < 0)
                            for token_id in word_ids
                        ):
                            typeerror()

                        constraint = PhrasalConstraint(word_ids)
                    final_constraints.append(constraint)

            # 11. prepare beam search scorer
            constrained_beam_scorer = ConstrainedBeamSearchScorer(
                constraints=final_constraints,
                batch_size=batch_size,
                num_beams=generation_config.num_beams,
                device=inputs_tensor.device,
                length_penalty=generation_config.length_penalty,
                do_early_stopping=generation_config.early_stopping,
                num_beam_hyps_to_keep=generation_config.num_return_sequences,
            )
            # 12. interleave input_ids with `num_beams` additional sequences per batch
            input_ids, model_kwargs = self._expand_inputs_for_generation(
                input_ids=input_ids,
                expand_size=generation_config.num_beams,
                is_encoder_decoder=self.config.is_encoder_decoder,
                **model_kwargs,
            )
            # 13. run beam search
            return self.constrained_beam_search(
                input_ids,
                constrained_beam_scorer=constrained_beam_scorer,
                logits_processor=logits_processor,
                stopping_criteria=stopping_criteria,
                pad_token_id=generation_config.pad_token_id,
                eos_token_id=generation_config.eos_token_id,
                output_scores=generation_config.output_scores,
                return_dict_in_generate=generation_config.return_dict_in_generate,
                synced_gpus=synced_gpus,
                **model_kwargs,
            )

    @torch.no_grad()
    def sample_stream(
        self,
        input_ids: torch.LongTensor,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        logits_warper: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[Union[int, List[int]]] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        **model_kwargs,
    ) -> Union[SampleOutput, torch.LongTensor]:
        r"""
        Generates sequences of token ids for models with a language modeling head using **multinomial sampling** and
        can be used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.

        <Tip warning={true}>

        In most cases, you do not need to call [`~generation.GenerationMixin.sample`] directly. Use generate() instead.
        For an overview of generation strategies and code examples, check the [following
        guide](./generation_strategies).

        </Tip>

        Parameters:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                The sequence used as a prompt for the generation.
            logits_processor (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
                used to modify the prediction scores of the language modeling head applied at each generation step.
            stopping_criteria (`StoppingCriteriaList`, *optional*):
                An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
                used to tell if the generation loop should stop.
            logits_warper (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsWarper`] used
                to warp the prediction score distribution of the language modeling head applied before multinomial
                sampling at each generation step.
            max_length (`int`, *optional*, defaults to 20):
                **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
                tokens. The maximum length of the sequence to be generated.
            pad_token_id (`int`, *optional*):
                The id of the *padding* token.
            eos_token_id (`int`, *optional*):
                The id of the *end-of-sequence* token.
            output_attentions (`bool`, *optional*, defaults to `False`):
                Whether or not to return the attentions tensors of all attention layers. See `attentions` under
                returned tensors for more details.
            output_hidden_states (`bool`, *optional*, defaults to `False`):
                Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
                for more details.
            output_scores (`bool`, *optional*, defaults to `False`):
                Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
            return_dict_in_generate (`bool`, *optional*, defaults to `False`):
                Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
            synced_gpus (`bool`, *optional*, defaults to `False`):
                Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
            model_kwargs:
                Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
                an encoder-decoder model the kwargs should include `encoder_outputs`.

        Return:
            [`~generation.SampleDecoderOnlyOutput`], [`~generation.SampleEncoderDecoderOutput`] or `torch.LongTensor`:
            A `torch.LongTensor` containing the generated tokens (default behaviour) or a
            [`~generation.SampleDecoderOnlyOutput`] if `model.config.is_encoder_decoder=False` and
            `return_dict_in_generate=True` or a [`~generation.SampleEncoderDecoderOutput`] if
            `model.config.is_encoder_decoder=True`.

        Examples:

        ```python
        >>> from transformers import (
        ...     AutoTokenizer,
        ...     AutoModelForCausalLM,
        ...     LogitsProcessorList,
        ...     MinLengthLogitsProcessor,
        ...     TopKLogitsWarper,
        ...     TemperatureLogitsWarper,
        ...     StoppingCriteriaList,
        ...     MaxLengthCriteria,
        ... )
        >>> import torch

        >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
        >>> model = AutoModelForCausalLM.from_pretrained("gpt2")

        >>> # set pad_token_id to eos_token_id because GPT2 does not have a EOS token
        >>> model.config.pad_token_id = model.config.eos_token_id
        >>> model.generation_config.pad_token_id = model.config.eos_token_id

        >>> input_prompt = "Today is a beautiful day, and"
        >>> input_ids = tokenizer(input_prompt, return_tensors="pt").input_ids

        >>> # instantiate logits processors
        >>> logits_processor = LogitsProcessorList(
        ...     [
        ...         MinLengthLogitsProcessor(15, eos_token_id=model.generation_config.eos_token_id),
        ...     ]
        ... )
        >>> # instantiate logits processors
        >>> logits_warper = LogitsProcessorList(
        ...     [
        ...         TopKLogitsWarper(50),
        ...         TemperatureLogitsWarper(0.7),
        ...     ]
        ... )

        >>> stopping_criteria = StoppingCriteriaList([MaxLengthCriteria(max_length=20)])

        >>> torch.manual_seed(0)  # doctest: +IGNORE_RESULT
        >>> outputs = model.sample(
        ...     input_ids,
        ...     logits_processor=logits_processor,
        ...     logits_warper=logits_warper,
        ...     stopping_criteria=stopping_criteria,
        ... )

        >>> tokenizer.batch_decode(outputs, skip_special_tokens=True)
        ['Today is a beautiful day, and a wonderful day.\n\nI was lucky enough to meet the']
        ```"""
        # init values
        logits_processor = (
            logits_processor if logits_processor is not None else LogitsProcessorList()
        )
        stopping_criteria = (
            stopping_criteria
            if stopping_criteria is not None
            else StoppingCriteriaList()
        )
        if max_length is not None:
            warnings.warn(
                "`max_length` is deprecated in this function, use"
                " `stopping_criteria=StoppingCriteriaList(MaxLengthCriteria(max_length=max_length))` instead.",
                UserWarning,
            )
            stopping_criteria = validate_stopping_criteria(
                stopping_criteria, max_length
            )
        logits_warper = (
            logits_warper if logits_warper is not None else LogitsProcessorList()
        )
        pad_token_id = (
            pad_token_id
            if pad_token_id is not None
            else self.generation_config.pad_token_id
        )
        eos_token_id = (
            eos_token_id
            if eos_token_id is not None
            else self.generation_config.eos_token_id
        )
        if isinstance(eos_token_id, int):
            eos_token_id = [eos_token_id]
        output_scores = (
            output_scores
            if output_scores is not None
            else self.generation_config.output_scores
        )
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.generation_config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.generation_config.output_hidden_states
        )
        return_dict_in_generate = (
            return_dict_in_generate
            if return_dict_in_generate is not None
            else self.generation_config.return_dict_in_generate
        )

        # init attention / hidden states / scores tuples
        scores = () if (return_dict_in_generate and output_scores) else None
        decoder_attentions = (
            () if (return_dict_in_generate and output_attentions) else None
        )
        cross_attentions = (
            () if (return_dict_in_generate and output_attentions) else None
        )
        decoder_hidden_states = (
            () if (return_dict_in_generate and output_hidden_states) else None
        )

        # keep track of which sequences are already finished
        unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)

        this_peer_finished = False  # used by synced_gpus only
        # auto-regressive generation
        while True:
            if synced_gpus:
                # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
                # The following logic allows an early break if all peers finished generating their sequence
                this_peer_finished_flag = torch.tensor(
                    0.0 if this_peer_finished else 1.0
                ).to(input_ids.device)
                # send 0.0 if we finished, 1.0 otherwise
                dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
                # did all peers finish? the reduced sum will be 0.0 then
                if this_peer_finished_flag.item() == 0.0:
                    break

            # prepare model inputs
            model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)

            # forward pass to get next token
            outputs = self(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
            )

            if synced_gpus and this_peer_finished:
                continue  # don't waste resources running the code we don't need

            next_token_logits = outputs.logits[:, -1, :]

            # pre-process distribution
            next_token_scores = logits_processor(input_ids, next_token_logits)
            next_token_scores = logits_warper(input_ids, next_token_scores)

            # Store scores, attentions and hidden_states when required
            if return_dict_in_generate:
                if output_scores:
                    scores += (next_token_scores,)
                if output_attentions:
                    decoder_attentions += (
                        (outputs.decoder_attentions,)
                        if self.config.is_encoder_decoder
                        else (outputs.attentions,)
                    )
                    if self.config.is_encoder_decoder:
                        cross_attentions += (outputs.cross_attentions,)

                if output_hidden_states:
                    decoder_hidden_states += (
                        (outputs.decoder_hidden_states,)
                        if self.config.is_encoder_decoder
                        else (outputs.hidden_states,)
                    )

            # sample
            probs = nn.functional.softmax(next_token_scores, dim=-1)
            next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1)

            # finished sentences should have their next token be a padding token
            if eos_token_id is not None:
                if pad_token_id is None:
                    raise ValueError(
                        "If `eos_token_id` is defined, make sure that `pad_token_id` is defined."
                    )
                next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
                    1 - unfinished_sequences
                )
            yield next_tokens
            # update generated ids, model inputs, and length for next step
            input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
            model_kwargs = self._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
            )

            # if eos_token was found in one sentence, set sentence to finished
            if eos_token_id is not None:
                unfinished_sequences = unfinished_sequences.mul(
                    (sum(next_tokens != i for i in eos_token_id)).long()
                )

            # stop when each sentence is finished, or if we exceed the maximum length
            if unfinished_sequences.max() == 0 or stopping_criteria(input_ids, scores):
                if not synced_gpus:
                    break
                else:
                    this_peer_finished = True


def init_stream_support():
    PreTrainedModel.generate = NewGenerationMixin.generate
    PreTrainedModel.sample_stream = NewGenerationMixin.sample_stream


if __name__ == "__main__":
    from transformers import PreTrainedModel
    from transformers import AutoTokenizer, AutoModelForCausalLM

    PreTrainedModel.generate = NewGenerationMixin.generate
    PreTrainedModel.sample_stream = NewGenerationMixin.sample_stream
    model = AutoModelForCausalLM.from_pretrained(
        "bigscience/bloom-560m", torch_dtype=torch.float16
    )

    tokenizer = AutoTokenizer.from_pretrained("bigscience/bloom-560m")
    model = model.to("cuda:0")
    model = model.eval()
    prompt_text = "hello? \n"
    input_ids = tokenizer(
        prompt_text, return_tensors="pt", add_special_tokens=False
    ).input_ids
    input_ids = input_ids.to("cuda:0")

    with torch.no_grad():
        result = model.generate(
            input_ids,
            max_new_tokens=200,
            do_sample=True,
            top_k=30,
            top_p=0.85,
            temperature=0.35,
            repetition_penalty=1.2,
            early_stopping=True,
            seed=0,
        )
        print(tokenizer.decode(result, skip_special_tokens=True))
        generator = model.generate(
            input_ids,
            max_new_tokens=200,
            do_sample=True,
            top_k=30,
            top_p=0.85,
            temperature=0.35,
            repetition_penalty=1.2,
            early_stopping=True,
            seed=0,
            do_stream=True,
        )
        stream_result = ""
        for x in generator:
            chunk = tokenizer.decode(x, skip_special_tokens=True)
            stream_result += chunk
        print(stream_result)
